import os

from hashlib import sha1

from sqlalchemy import ForeignKey, Column, types, Table
from sqlalchemy.orm import relationship, backref

# imports from the model
from abstrackr.model.meta import Session, Base

# authentication
#from abstrackr.model.auth import User, Group, Permission


def init_model(engine):
    """Call me before using any of the tables or classes in the model"""
    Session.configure(bind=engine)


### Association Tables
citations_tasks_table = Table('citations_tasks', Base.metadata,
        Column('citation_id', types.Integer, ForeignKey('citations.id')),
        Column('task_id', types.Integer, ForeignKey('tasks.id'))
)

users_projects_table = Table('users_projects', Base.metadata,
        Column('user_id', types.Integer, ForeignKey('user.id')),
        Column('project_id', types.Integer, ForeignKey('projects.id'))
)

projects_leaders_table = Table('projects_leaders', Base.metadata,
        Column('user_id', types.Integer, ForeignKey('user.id')),
        Column('project_id', types.Integer, ForeignKey('projects.id'))
)

### end of Association Tables

class Project(Base):
    __tablename__ = "projects"
    #__mapper_args__ = dict(order_by="date desc")

    id = Column(types.Integer, primary_key=True)
    #leader_id = Column(types.Integer, ForeignKey('user.id'))

    # If initial_round_size > 0, this will point to
    # the entry in the citations_tasks table that
    # maps to the initial assignment associated with
    # this project.
    initial_assignment_id = Column(types.Integer)

    name = Column(types.Unicode(255))
    description = Column(types.Unicode(10000))

    # This is used to identify the project when
    # the leader invites others to the project
    code = Column(types.Unicode(10), index=True, unique=True)

    # `single', `double', or `advanced'
    screening_mode = Column(types.Unicode(50))

    # True := tags are private
    # False := tags are public
    tag_privacy = Column(types.Boolean)

    # the number of labels to be procured for each abstract
    num_labels_thus_far = Column(types.Integer)

    # basically, the AI criteria (or random)
    sort_by = Column(types.Unicode(255))

    # If >0, this represents a fixed set of
    # citations that will (potentially) be
    # screened by everyone on the project
    initial_round_size = Column(types.Integer)

    # Minimum and maximum number of citations a user should screen for this project
    min_citations = Column(types.Integer)
    max_citations = Column(types.Integer)

    # Bookkeeping
    date_created = Column(types.DateTime())
    date_modified = Column(types.DateTime())

    priorities = relationship('Priority', order_by='Priority.id', backref='project')
    citations = relationship('Citation', order_by='Citation.id', backref='project')
    assignments = relationship('Assignment', order_by='Assignment.id', backref='project')
    labels = relationship('Label', order_by='Label.id', backref='project')
    members = relationship('User', secondary=users_projects_table, backref='member_of_projects')
    leaders = relationship('User', secondary=projects_leaders_table, backref='leader_of_projects')


class Citation(Base):
    __tablename__ = "citations"

    # note that this is independent of pubmed/refman/whatever id!
    id = Column(types.Integer, primary_key=True)

    # associates the citation with the project that owns it
    project_id = Column(types.Integer, ForeignKey('projects.id'))
    pmid = Column(types.Integer)
    refman = Column(types.Integer)

    title = Column(types.Unicode(500))
    # length is based on back-of-envelope calculation
    abstract = Column(types.Unicode(10000))
    authors = Column(types.Unicode(5000))
    journal = Column(types.Unicode(1000))
    publication_date = Column(types.DateTime())
    keywords = Column(types.Unicode(5000))

    tasks = relationship("Task", secondary=citations_tasks_table, backref="citations")
    priorities = relationship('Priority', backref="citation")
    labels = relationship("Label", backref="citation")

class Priority(Base):
    """
    This table stores the ordered priority for citation screening, i.e.,
    facilitates active learning. It also attempts to solve the potential
    problem of users labeling the same citation simultaneously by
    containing a `is_out` field and date.

    """

    __tablename__ = "priorities"

    id = Column(types.Integer, primary_key=True)

    # ForeignKey relationship columns
    project_id = Column(types.Integer, ForeignKey('projects.id'))
    citation_id = Column(types.Integer, ForeignKey('citations.id'))

    priority = Column(types.Integer)

    # we keep the number of times that each citation has been labeled, and
    # remove it from the queue when a sufficient number of labels have been
    # collected
    num_times_labeled = Column(types.Integer)

    # Keep a record of whether this priority item is currently being labeled
    # by a user. This is an artificial lock on the item so that users do not
    # overwrite each other's label
    is_out = Column(types.Boolean)
    locked_by = Column(types.Integer)
    time_requested = Column(types.DateTime())


class TagType(Base):
    """User added tags"""

    __tablename__ = "tagtypes"

    id = Column(types.Integer, primary_key=True)
    text = Column(types.Unicode(500))
    project_id = Column(types.Integer)
    #label = Column(types.SmallInteger)
    creator_id = Column(types.Integer)
    color = Column(types.Unicode(50))


class Tag(Base):
    """Stores study/tag pairs"""

    __tablename__ = "tags"

    id = Column(types.Integer, primary_key=True)
    tag_id = Column(types.Integer)
    creator_id = Column(types.Integer)
    citation_id = Column(types.Integer)


class Note(Base):
    """Stores notes; both structured and unstructured"""

    __tablename__ = "notes"

    id = Column(types.Integer, primary_key=True)
    creator_id = Column(types.Integer)
    citation_id = Column(types.Integer)
    general = Column(types.Unicode(1000))
    population = Column(types.Unicode(1000))
    ic = Column(types.Unicode(1000)) # intervention/comparator
    outcome = Column(types.Unicode(1000))


class LabeledFeature(Base):
    """Stores labeled features, i.e., terms"""

    __tablename__ = "labeledfeatures"

    id = Column(types.Integer, primary_key=True)
    term = Column(types.Unicode(500))
    # project for which this term applies
    project_id = Column(types.Integer)
    # person who entered the term
    user_id = Column(types.Integer)
    # label
    label = Column(types.SmallInteger)
    date_created = Column(types.DateTime())


class Label(Base):
    """Stores instances labels"""

    __tablename__ = "labels"

    id = Column(types.Integer, primary_key=True)
    # project for which this document was screened
    project_id = Column(types.Integer, ForeignKey('projects.id'))
    study_id = Column(types.Integer, ForeignKey('citations.id')) # TODO: need to rename this to citation_id
    user_id = Column(types.Integer)
    assignment_id = Column(types.Integer)
    # -1, 0, 1
    label = Column(types.SmallInteger)
    # in seconds
    labeling_time = Column(types.Integer)
    first_labeled = Column(types.DateTime())
    label_last_updated = Column(types.DateTime())


class Assignment(Base):

    __tablename__ = "assignments"

    id = Column(types.Integer, primary_key=True)
    project_id = Column(types.Integer, ForeignKey('projects.id'))
    user_id = Column(types.Integer, ForeignKey('user.id'))
    task_id = Column(types.Integer, ForeignKey('tasks.id'))
    done_so_far = Column(types.Integer)
    date_assigned = Column(types.DateTime())
    date_due = Column(types.DateTime())
    done = Column(types.Boolean)

    # both of these fields are redundant
    # with the corresponding Task entry,
    # but it is easier to keep them here, too.
    num_assigned = Column(types.Integer)
    # this is the same as `task_type'.
    assignment_type = Column(types.Unicode(50))


class Task(Base):
    """
    a Task is a unit of work. Tasks have types;
    some are `perpetual`, i.e., they basically say `allow the
    person to whom this is assigned to keep on labeling, until
    there are no more citations that have not been labeled the
    desired number of times`, and there are `finite`, in which
    assignees are to label a fixed number of citations. Tasks are
    always associated with exactly one project, but multiple
    users can be assigned the same Task (see the Assignment table).

    """

    __tablename__ = "tasks"

    id = Column(types.Integer, primary_key=True)
    project_id = Column(types.Integer)
    # this is 'perpetual', 'initial' or 'finite'; the former indicates a 'perpetual'
    # screening task -- i.e., they will continue screening
    # while abstracts remain in the Priority queue for this
    # project. The latter two are special cases; in which only n
    # citations will be screened.
    task_type = Column(types.Unicode(50))
    # both of the following are N/A for 'perpetual'
    num_assigned = Column(types.Integer)

    assignments = relationship("Assignment", order_by='Assignment.id', backref="task")


class EncodeStatus(Base):
    """
    This table contains one entry for each dataset, indicating
    its encoded (i.e., for curious_snake/machine learning stuff)
    status. That is, this table contains information regarding:
        1) if the dataset has been encoded
        2) when the labels of this encoded file were last updated
        3) the path to the encoded file

    """

    __tablename__ = "encodedstatuses"

    id = Column(types.Integer, primary_key=True)
    project_id = Column(types.Integer) # associated project
    is_encoded = Column(types.Boolean) # has it been encoded yet?
    labels_last_updated = Column(types.DateTime())
    # the location of the base directory for the encoded project
    base_path = Column(types.Unicode(100))


class PredictionsStatus(Base):
    """Status of predictions (do they exist? last update, etc.)"""

    __tablename__ = "predictionstatuses"

    id = Column(types.Integer, primary_key=True)
    project_id = Column(types.Integer) # associated project
    predictions_exist = Column(types.Boolean) # has it been encoded yet?
    predictions_last_made = Column(types.DateTime())
    train_set_size = Column(types.Integer) # how many did we train on?
    num_pos_train = Column(types.Integer) # number of positive examples we trained on


class Prediction(Base):
    """Current inclusion/exclusion predictions for studies"""

    __tablename__ = "predictions"

    id = Column(types.Integer, primary_key=True)
    study_id = Column(types.Integer) # the (internal) study id
    project_id = Column(types.Integer) # it makes life easier to have this around
    prediction = Column(types.Boolean) # true = include; false = exclude
    num_yes_votes = Column(types.Float) # number of ensemble members that voted yes
    predicted_probability = Column(types.Float) # predicted probability

    


####################################
## these tables for authentication #
####################################
class ResetPassword(Base):
    """
    This table facilitates password recovery. If someone requests to reset
    their password, we generate a random token and insert it into this db.
    We then email them this token, verifying their id.

    """

    __tablename__ = "ResetPassword"

    id = Column(types.Integer, primary_key=True)
    user_email = Column(types.Unicode(80))
    token = Column(types.Unicode(10))

class Group(Base):
    __tablename__ = "group"
    id  = Column(types.Integer, primary_key=True)
    name = Column(types.Unicode(255))

class Permission(Base):
    __tablename__ = "permission"
    id  = Column(types.Integer, primary_key=True)
    name = Column(types.Unicode(255))

class User(Base):

    __tablename__ = "user"

    id  = Column(types.Integer, primary_key=True)
    username = Column(types.Unicode(255))
    email = Column(types.Unicode(80))
    password = Column(types.Unicode(80))
    fullname = Column(types.Unicode(255))
    # Number of Systematic Reviews they've been involved with
    experience = Column(types.Integer)
    # These three columns are meant to hold user's preference
    # choices as to whether to show the corresponding information
    # on the citation screen (default=True)
    show_journal = Column(types.Boolean, default=True)
    show_authors = Column(types.Boolean, default=True)
    show_keywords = Column(types.Boolean, default=True)

    assignments = relationship("Assignment", backref=backref('user', order_by=id))

    def _set_password(self, password):
        """Hash password on the fly."""

        hashed_password = password

        if isinstance(password, unicode):
            password_8bit = password.encode('UTF-8')
        else:
            password_8bit = password

        salt = sha1()
        salt.update(os.urandom(60))
        hash = sha1()
        hash.update(password_8bit + salt.hexdigest())
        hashed_password = salt.hexdigest() + hash.hexdigest()

        # Make sure the hased password is an UTF-8 object at the end of the
        # process because SQLAlchemy _wants_ a unicode object for Unicode
        # fields
        if not isinstance(hashed_password, unicode):
            hashed_password = hashed_password.decode('UTF-8')

        self.password = hashed_password

    def _get_password(self):
        """Return the password hashed"""

        return self.password

    def validate_password(self, password):
        """
        Check the password against existing credentials.

        :param password: the password that was provided by the user to
            try and authenticate. This is the clear text version that we will
            need to match against the hashed one in the database.
        :type password: unicode object.
        :return: Whether the password is valid.
        :rtype: bool

        """

        hashed_pass = sha1()
        hashed_pass.update(password + self.password[:40])
        return self.password[40:] == hashed_pass.hexdigest()


# This is the association table for the many-to-many relationship between
# groups and permissions.
class GroupPermission(Base):
    __tablename__ = "group_permission"
    id  = Column(types.Integer, primary_key=True)
    group_id = Column(types.Integer)
    permission_id = Column(types.Integer)

# This is the association table for the many-to-many relationship between
# groups and users
class UserGroup(Base):
    __tablename__ = "user_group"
    id = Column(types.Integer, primary_key=True)
    user_id = Column(types.Integer)
    group_id = Column(types.Integer)



"""
SQLAlchemy-powered model definitions for repoze.what SQL plugin.
Sets up Users, Groups and Permissions
"""
from sqlalchemy import Table, ForeignKey, Column
from sqlalchemy.types import Unicode, UnicodeText, Integer, Date, CHAR
from sqlalchemy import orm
from abstrackr.model.meta import metadata
import os
from hashlib import sha1

group_table = Table('group', metadata,
    Column('id', Integer(), primary_key=True),
    Column('name', Unicode(255), unique=True, nullable=False),
)

permission_table = Table('permission', metadata,
    Column('id', Integer(), primary_key=True),
    Column('name', Unicode(255), unique=True, nullable=False),
)

user_table = Table('user', metadata,
    Column('id', Integer(), primary_key=True),
    Column('username', Unicode(255), unique=True, nullable=False),
    Column('email', Unicode(255), unique=True, nullable=False),
    Column('password', Unicode(80), nullable=False),
    Column('fullname', Unicode(255), nullable=False),
)

# This is the association table for the many-to-many relationship between
# groups and permissions.
group_permission_table = Table('group_permission', metadata,
    Column('group_id', Integer, ForeignKey('group.id')),
    Column('permission_id', Integer, ForeignKey('permission.id')),
)

# This is the association table for the many-to-many relationship between
# groups and users
user_group_table = Table('user_group', metadata,
    Column('user_id', Integer, ForeignKey('user.id')),
    Column('group_id', Integer, ForeignKey('group.id')),
)

class Group(object):
    pass

class Permission(object):
    pass

class User(object):

    def _set_password(self, password):
        """Hash password on the fly."""
        hashed_password = password

        if isinstance(password, unicode):
            password_8bit = password.encode('UTF-8')
        else:
            password_8bit = password

        salt = sha1()
        salt.update(os.urandom(60))
        hash = sha1()
        hash.update(password_8bit + salt.hexdigest())
        hashed_password = salt.hexdigest() + hash.hexdigest()

        # Make sure the hased password is an UTF-8 object at the end of the
        # process because SQLAlchemy _wants_ a unicode object for Unicode
        # fields
        if not isinstance(hashed_password, unicode):
            hashed_password = hashed_password.decode('UTF-8')

        self.password = hashed_password

    def _get_password(self):
        """Return the password hashed"""
        return self.password

    def validate_password(self, password):
        """
        Check the password against existing credentials.

        :param password: the password that was provided by the user to
            try and authenticate. This is the clear text version that we will
            need to match against the hashed one in the database.
        :type password: unicode object.
        :return: Whether the password is valid.
        :rtype: bool

        """
        hashed_pass = sha1()
        hashed_pass.update(password + self.password[:40])
        return self.password[40:] == hashed_pass.hexdigest()

# Map SQLAlchemy table definitions to python classes
orm.mapper(Group, group_table, properties={
    'permissions':orm.relation(Permission, secondary=group_permission_table),
    'users':orm.relation(User, secondary=user_group_table),
})
orm.mapper(Permission, permission_table, properties={
    'groups':orm.relation(Group, secondary=group_permission_table),
})
orm.mapper(User, user_table, properties={
    'groups':orm.relation(Group, secondary=user_group_table),
})



import formencode
from formencode import validators
from pylons.decorators import validate
import abstrackr.model as model

class UniqueUsername(formencode.FancyValidator):
    def _to_python(self, value, state):
        user_q = model.meta.Session.query(model.User)
        usernames = [user.username for user in user_q.all()]
        if value in usernames:
            raise formencode.Invalid(
                    'Sorry, that username already exists. Try something else.',
                     value, state)
            return value
            
class RegisterForm(formencode.Schema):
    allow_extra_fields = True
    filter_extra_fields = True
    email = formencode.validators.Email(not_empty=True)
    first_name = formencode.validators.String(not_empty=True)
    last_name =  formencode.validators.String(not_empty=True)
    experience = formencode.validators.Int(not_empty=True)
    password = formencode.validators.NotEmpty()
    username = UniqueUsername()
    

    
class ChangePasswordForm(formencode.Schema):
    allow_extra_fields = True
    chained_validators = [validators.FieldsMatch(
                         'password', 'password_confirm')]





#!/usr/bin/env python
# encoding: utf-8
#
# 
# The MIT License
# 
# Copyright (c) 2009 Byron Wallace
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

'''
Byron C Wallace
Tufts Medical Center

A module for encoding text documents. Used to be just tf-idf; has since grown
to include domain adaptation tricks, &etc.

Assumes you have cleaned individual title files in "Titles" directory
It is also assumed that the names of these files (e.g,. 1.*) map
to an ID (e.g., refman ID). I.e., 1.* -> reference id 1. These are used as identifiers.

Usage (assuming you've a directory of text files "Titles"')
>titles = tfidf2.build_bag_of_words_over_dir("Titles")

You'll need a list of ids that are 'positives' (e.g., "relevant")
>pos_indices = { read in and build this list }

Now we can dump to some file specified by outpath
>tfidf_to_file_for_lib_SVM(titles, pos_indices, outpath)

Alternatively, do it all at once with the encode_docs method.
'''

import re
import math
import string
import os
import pdb
import csv
import sys
import random
from operator import itemgetter
import pickle

###
# 5/6/11
# if you want to do stuff with author networks, you need
# to import the networks module. however, this leads to a
# circular import, which is problematic...
#import networks
try:
  import nose
except:
  print "nose isn't installed -- can't run unittests!"

### 
# 9/20/11
# required stuff for SCL, &etc.
###
try:
  import numpy
  import liblinear
  import liblinearutil
  from liblinear import *
  from liblinearutil import *
except:
  print "either numpy or liblinear isn't installed -- won't be able to do fancy stuff like SCL"

import copy


#stop_list_path=os.getcwd() + "/abstrackr/lib/stop_list.txt"
stop_list_path = "/Users/abstrackr-user/Hive/abstrackr/abstrackr/lib/stop_list.txt"

print "stop word list path is %s" % stop_list_path

def build_stop_list(stop_list_path):
    exclude_words = ["a", "about", "above", "across", "after", "afterwards", "again", "against", "all", "almost", "alone", "along", "already", "also", "although", "always", "am", "among", "amongst", "amoungst", "amount", "an", "and", "another", "any", "anyhow", "anyone", "anything", "anyway", "anywhere", "are", "around", "as", "at", "back", "be", "became", "because", "become", "becomes", "becoming", "been", "before", "beforehand", "behind", "being", "below", "beside", "besides", "between", "beyond", "bill", "both", "bottom", "but", "by", "call", "can", "cannot", "cant", "co", "con", "could", "couldnt", "cry", "de", "describe", "detail", "do", "done", "down", "due", "during", "each", "eg", "eight", "either", "eleven", "else", "elsewhere", "empty", "enough", "etc", "even", "ever", "every", "everyone", "everything", "everywhere", "except", "few", "fifteen", "fify", "fill", "find", "fire", "first", "five", "for", "former", "formerly", "forty", "found", "four", "from", "front", "full", "further", "get", "give", "go", "had", "has", "hasnt", "have", "he", "hence", "her", "here", "hereafter", "hereby", "herein", "hereupon", "hers", "herself", "him", "himself", "his", "how", "however", "hundred", "i", "ie", "if", "in", "inc", "indeed", "interest", "into", "is", "it", "its", "itself", "keep", "last", "latter", "latterly", "least", "less", "ltd", "made", "many", "may", "me", "meanwhile", "might", "mill", "mine", "more", "moreover", "most", "mostly", "move", "much", "must", "my", "myself", "name", "namely", "neither", "never", "nevertheless", "next", "nine", "no", "nobody", "none", "noone", "nor", "not", "nothing", "now", "nowhere", "of", "off", "often", "on", "once", "one", "only", "onto", "or", "other", "others", "otherwise", "our", "ours", "ourselves", "out", "over", "own", "part", "per", "perhaps", "please", "put", "rather", "re", "same", "see", "seem", "seemed", "seeming", "seems", "serious", "several", "she", "should", "show", "side", "since", "sincere", "six", "sixty", "so", "some", "somehow", "someone", "something", "sometime", "sometimes", "somewhere", "still", "such", "system", "take", "ten", "than", "that", "the", "their", "them", "themselves", "then", "thence", "there", "thereafter", "thereby", "therefore", "therein", "thereupon", "these", "they", "thick", "thin", "third", "this", "those", "though", "three", "through", "throughout", "thru", "thus", "to", "together", "too", "top", "toward", "towards", "twelve", "twenty", "two", "un", "under", "until", "up", "upon", "us", "very", "via", "was", "we", "well", "were", "what", "whatever", "when", "whence", "whenever", "where", "whereafter", "whereas", "whereby", "wherein", "whereupon", "wherever", "whether", "which", "while", "whither", "who", "whoever", "whole", "whom", "whose", "why", "will", "with", "within", "without", "would", "yet", "you", "your", "yours", "yourself", "yourselves"]
#    exclude_words = []
#    if stop_list_path != None:
#        f = open(stop_list_path)
#        while 1:
#            line = f.readline()
#            if not line:
#                break
#            # Every line is assumed to be a single word
#            exclude_words.append(line.strip())
    return exclude_words

# if you don't want to use a stop list, just set the stop_list global
# to an empty list (stop_list = [])
print "building stop word list..."
stop_list = build_stop_list(stop_list_path)
print "done."


def add_labels(new_path, lbl_dict):
    lbls = [x for x in csv.reader(open(new_path, 'r'))]
    for l in lbls:
        #lbl_dict[eval(l[0])] = "-1" if l[1]=="n" else "1"
        lbl_dict[eval(l[0])] = "1" if l[1] in ("y", "m") else "-1"
    return lbl_dict
    

def update_labels(fpath, outpath, new_lbl_d, unlabel_if_not_in_dict=False, ignore_these=[]):
    ''' relabels the instances in fpath with new_lbl_d; writes to outpath. '''
    if not isinstance(new_lbl_d.keys()[0] , int):
        new_lbl_d = lbls_to_strs(new_lbl_d)
        
    new_file = []
    lines = open(fpath).readlines()
    for line in lines:
        line_sp = line.split(" ")
        cur_id = eval(line_sp[0])

        if cur_id not in ignore_these:
            if cur_id in new_lbl_d.keys():
                line_sp[1] = str(new_lbl_d[cur_id])
            elif unlabel_if_not_in_dict:
                line_sp[1] = "?"
            try:
                new_file.append(" ".join(line_sp))
            except:
                pdb.set_trace()
    outf = open(outpath, 'w')
    outf.write("".join(new_file))
    

def lbls_to_strs(lbl_d):
    new_d = {}
    for xid in lbl_d:
        new_d[xid] = str(lbl_d[xid])
    return new_d

def abstrackr_encode(dir_path, out_dir, label_d=None, \
                        fields=["AB", "MH", "TI"], min_word_count=3,\
                        bi_grams_too=False):
    cleaned_dir_d = {}
    for field in fields:
        field_path = os.path.join(dir_path, field)
        clean_docs_path = os.path.join(field_path, "Cleaned")
        if not os.path.exists(clean_docs_path):
            _mkdir(clean_docs_path)
        clean_up_docs(field_path, out_dir=clean_docs_path)
        cleaned_dir_d[field] = clean_docs_path
    
    merge_fields(cleaned_dir_d, out_dir)

    print "ok -- fields cleaned and written out. building BOW"

    encode_docs(out_dir, os.path.join(out_dir, "encoded"), "merged_encoded", \
                lbl_dict=label_d, clean_first=False, bi_grams_too=bi_grams_too,\
                min_word_count=min_word_count)

def encode_docs(dir_path, out_path, out_f_name, 
                lbl_dict=None, clean_first=True, binary=True, 
                min_word_count=3, bi_grams_too=False, delete_files_after=False):
    '''
    Given a directory path, this method cleans, then encodes and writes out all text files therein.
    '''
    
    if lbl_dict is not None and len(lbl_dict)>0:
        if not isinstance(lbl_dict.values()[0], str):
            print "\n\n\n*****labels are *not* strings, but they need to be!\n going to force this, MAKE SURE THIS IS SANE\n\n"
            lbl_dict = lbls_to_strs(lbl_dict)
 
            
    # first, clean the documents
    if clean_first:
        print "cleaning documents.."
        clean_docs_path = os.path.join(dir_path, "Cleaned")
        if not os.path.exists(clean_docs_path):
            _mkdir(clean_docs_path)
        clean_up_docs(dir_path, out_dir=clean_docs_path)
        print "done cleaning."
    else:
        clean_docs_path = dir_path

    if not os.path.exists(out_path):
        _mkdir(out_path)

    # now build tf/idf representation
    print "encoding..."
    encoded_docs = build_bag_of_words_over_dir(
          clean_docs_path, binary_encode=binary, 
          min_word_count=min_word_count, bi_grams_too=bi_grams_too,
          word_index_path=os.path.join(out_path, "word_index.txt"))
    print "done encoding."
    
    # now write it out
    print "writing doc out..."
    pos_ids = []
    neg_ids = []      
    if lbl_dict is not None:
        ### note that we count 0s as positive!
        pos_ids = [str(x_id) for x_id in lbl_dict.keys() if lbl_dict[x_id] in ["1", "1.0", "0"]]
        neg_ids = [str(x_id) for x_id in lbl_dict.keys() if not lbl_dict[x_id] in ["1", "1.0"]]
     

    tfidf_to_file_for_lib_SVM_multi_label(encoded_docs, pos_ids, neg_ids,
                                                            None, os.path.join(out_path, out_f_name))
    
    if delete_files_after:
        # then we wipe out the files from disk
        print "removing files from disk..."
        # remove the original files
        clean_up(dir_path)
        # and the 'clean' files
        clean_up(clean_docs_path)
        print "all clean."

    print "done."
    
def clean_up(dir_path, word_file_name="word_index.txt"):
    
    #word_index_path = os.path.join(dir_path, "word_index.txt")

    files_to_del = [f for f in os.listdir(dir_path) 
                        if f != word_file_name and 
                        not os.path.isdir(os.path.join(dir_path, f))]

    for p in files_to_del:
        full_file_path = os.path.join(dir_path, p)
        try:
            os.remove(full_file_path)
        except:
            print "failed to delete file %s!" % p


def bag_of_authors(ids_to_authors):
    all_authors = []
    for author_list in ids_to_authors.values():
        all_authors.extend(author_list)
    
    authors_set = list(set(all_authors))
    # filter list
    authors_set = [author for author in authors_set if all_authors.count(author)>=3]
    author_indices_d = {}
    for i, author in enumerate(authors_set):
        author_indices_d[author] = i
    
    # now walk over all ids
    bag_o_authors = {}
    for article_id in ids_to_authors.keys():
        bag_o_authors[article_id] = {}
        for author in [article_author for article_author in ids_to_authors[article_id] if article_author in authors_set]:
            bag_o_authors[article_id][author_indices_d[author]] = 1.0
            
    return bag_o_authors
        
def xml_to_bag_of_authors(xml_path, lbl_d, out_path):
    ids_to_authors = networks.get_ids_to_authors(xml_path)
    assert(lbl_d.keys() == ids_to_authors.keys()) # although ordering doesn't matter...
    boa = bag_of_authors(ids_to_authors)
    out_str = []
    for xid in ids_to_authors.keys():
        #id, level1_lbl, level2_lbl, x
        out_str.append(lib_svm_str_multi_label(xid, lbl_d[xid], None, boa[xid]))
    fout = open(out_path, 'w')
    fout.write("\n".join(out_str))
    

def remove_docs(f_path, ids_to_remove):
    pass
    

def tdidf(wordfreqs, freqvecs):
    '''
    Returns tf-idf feature vectors. For a simple explanation, see: http://instruct.uwo.ca/gplis/601/week3/tfidf.html
  
    wordfreqs -- Vector s.t. w[i] is the total number of times word i was seen over all documents.
    freqvecs -- A dictionary mapping document ids to (raw) frequency vectors. 
    
    returns a dictionary mapping the keys in freqvecs to their tf-idf feature-vector representation
    '''
    N = len(freqvecs.keys()) # Total number of documents
    num_terms = len(wordfreqs) # Number of terms
    print "Number of documents: %s, number of terms %s" % (N, num_terms)
    print "Building n_vec..."
  
    # i is the document index; j the word/term index
    print "Building n_vec..."
    n_vec = [0 for j in range(num_terms)]
    for i in range(N):
      cur_doc = freqvecs[freqvecs.keys()[i]]
      for j in range(num_terms):
          if cur_doc[j] > 0:
              n_vec[j]+=1

    print "Word counts built, Now constructing TDF vector."
    tdfvecs = {}
    last_key = None
    for i in range(N):
      cur_key = freqvecs.keys()[i]
      last_key = cur_key
      
      if i%100 == 0:
          print "On document %s" % i
  
      cur_doc = freqvecs[cur_key]
      tdfvec = [0 for k in range(num_terms)] 

      for j in range(num_terms):
          # tf * idf
          tdfvec[j] = cur_doc[j] * math.log(N/n_vec[j], 2.0)
      # Normalize by the l2 norm
      cos_norm = math.sqrt(sum([tdfvec[j]**2 for j in range(num_terms)]))
      if cos_norm == 0:
          # None of the terms were in this document. Just return a vector of zeros.
          tdfvec = [0 for i in range(num_terms)]
      else:
          tdfvec = [tdfvec[j]/cos_norm for j in range(num_terms)]
  
      if cur_key in tdfvecs:
        print "Error -- key (doc id) already exists???"
        pdb.set_trace()
        
    
      tdfvecs[cur_key]=tdfvec
    return tdfvecs
  

def build_word_count_vector_for_doc(words, doc):
    '''
    Returns a vector V where V_i corresponds to the number of times words_i is contained in doc
    '''
    ### TODO implement sparse encoding here!
    count_vec = [doc.count(word) for word in words]
    return count_vec

    
def build_bag_of_words_feature_vectors(ids_to_texts, words, binary_encode=False):
    freq_vecs = {}
    for id in ids_to_texts.keys():
        if not binary_encode:
            freq_vecs[id] = build_word_count_vector_for_doc(words, ids_to_texts[id])
        else:
            # binary encoding -- 1or 0 for each word (present or not, respectively)
            
            ####
            # 12/17/10 -- sparse encoding now!
            ####
            doc = ids_to_texts[id]
            freq_vecs[id] = {}
            for i in xrange(len(words)):
                word = words[i]
                if word in doc:
                    freq_vecs[id][i] = 1.0
	
    if binary_encode:
        return freq_vecs
        
    word_freqs = [sum([ids_to_texts[id].count(w) for id in ids_to_texts.keys()]) for w in words]
    return tdidf(word_freqs, freq_vecs)

def build_bag_of_words_feature_vectors_scaled(ids_to_texts, words, labeled_terms, c=2):
    #
    # !!! Not finished. May want to implement this. 
    # Idea: scale entries in feature vectors to weight 
    # labeled terms more heavily than the rest (see Raghaven et al)
    #
    
    freq_vecs = {}
    for id in ids_to_texts.keys():
        # binary encoding -- 1or 0 for each word (present or not, respectively)
        doc = ids_to_texts[id]
        for word in words:
            cur_fv = []
            if word in doc:
                if word in labeled_terms:
                    cur_fv.append(c)
                else:
                    cur_fv.append(1.0)
            else:
                cur_fv.append(0.0)
                
        freq_vecs[id] = [1.0 if word in doc else 0.0 for word in words]

    return freq_vecs
    
    
    
def build_bag_of_words_over_dir(dir_path, split_txt_on = " ", binary_encode = False, \
                                word_index_path = None, word_index_out_dir=None,\
                                min_word_count = 3, bi_grams_too=False):
    '''
    Build bag of words representation vectors over *all of the documents* in dir_path.
    Note that we assume the documents are already clean.
    '''
    
    if word_index_path is None:
        word_index_path = os.path.join(dir_path, "word_index.txt")
    
    # read all the words in     
    s_words = []
    files_in_dir = [f for f in os.listdir(dir_path) if not os.path.isdir(os.path.join(dir_path, f))]

    for p in files_in_dir:
      try:
          doc_text = open(os.path.join(dir_path, p), 'r').readlines()[0].split(" ")
          s_words.extend(doc_text)
          if bi_grams_too:
            # then append bi-grams
            bi_grams = extract_bi_grams(doc_text)
            s_words.extend(bi_grams)
      except:
          pass
    
    unique_word_dict = {}
    set_words = list(set(s_words))
    for w in set_words:
        unique_word_dict[w] = 0
    
    ids_to_txt = {}
    words = []
    # ignore the word_index.txt file, which we generated
    for p in [f for f in files_in_dir if not "word_index.txt" in f]:
        cur_txt =[""]

        try:
            cur_txt = open(os.path.join(dir_path, p), 'r').readline().split(split_txt_on)

            if bi_grams_too:
                cur_txt.extend(extract_bi_grams(cur_txt))

            for word in cur_txt:
                unique_word_dict[word] += 1
        

        except  Exception, e:
            # abstract is missing!
            pass
        
        id = p.split(".")[0]
        ids_to_txt[id] = cur_txt
        words.extend(cur_txt)
    
    print "number of words: %s; number of unique words: %s" % (len(words), len(set_words))
    words = [word for word in unique_word_dict.keys() if unique_word_dict[word] >= min_word_count]

    word_index_out = open(word_index_path, 'w')
    word_index_out.write(str(words))
    word_index_out.close()

    return build_bag_of_words_feature_vectors (ids_to_txt, words, binary_encode=binary_encode)

def extract_bi_grams(words):
    # assumes words is an *ordered* list of strings
    bi_grams = []
    for w_i in xrange(len(words)-1):
        bi_grams.append("%s %s" % (words[w_i], words[w_i+1]))
    return bi_grams

def clean_up_txt(doc):
    ''' 
    Cleans and returns the parametric abstract text. I.e., strips punctuation, etc. Also removes
    any words in the stop list (if provided).
    '''
    words = []
    exclude_words = []
            
    # for hyphenated words, it makes more sense to split the
    # atoms and append both parts to the word list.
    split_on = ["-", "/"]
    for sp_char in split_on:
        for word in doc: 
            if sp_char in word:
                doc.extend(word.split(sp_char))
                try:
                    doc.remove(word)
                except:
                    pdb.set_trace()
    
    exclude = set(string.punctuation)
    for word in doc:
        word = word.lower()
        clean_word = ''.join(c for c in word if c not in exclude).strip()
        if clean_word and not clean_word in stop_list:
            if isnumerical(clean_word):
                clean_word = "NUMBER"
            words.append(clean_word)

    return words

def isnumerical(w):
    try:
        xfl = float(w)
        return True
    except:
        return False
    
    
def clean_up_docs(dir_path, out_dir = None, overwrite_dirty=False, grammar_only=False):
    if out_dir is not None:
        try:
            _mkdir(out_dir)
        except:
            pass # presumably the directory already exists
    else:
        out_dir = dir_path
    
    print "\ncleaning documents in %s..." % dir_path
    for doc in [f for f in os.listdir(dir_path) if not os.path.isdir(os.path.join(dir_path, f))]:
        dirty_path = os.path.join(dir_path, doc)
        ###
        # NOTE: we used to just read the first line and split on the
        #dirty_doc = open(dirty_path, 'r').readline().split(" ")
        all_text = " ".join(open(dirty_path, 'r').readlines())
        dirty_doc = all_text.split(" ")
        
        if not grammar_only:
            clean_doc = clean_up_txt(dirty_doc)
        else:
            # just stripping commas
            clean_doc = [x.replace(",", "") for x in dirty_doc]

        out_path = dirty_path if overwrite_dirty else dirty_path + ".cleaned"
        if out_dir != dir_path:
            # then an output directory was passed in
            out_path = os.path.join(out_dir, doc)
        clean_doc_out = open(out_path, 'w')
        clean_doc_out.write(" ".join(clean_doc))
    print "documents cleaned and written."


def merge_fields(cleaned_dirs_d, out_dir):
    '''
    cleaned_dirs_d -- dictionary mapping fields to directories containing
                        (cleaned) files. e.g., "abstract" -> "my_path/abstracts/Cleaned"
    '''
    try:
        _mkdir(out_dir)
    except:
        pass # probably already exists


    for file_id in [f for f in os.listdir(cleaned_dirs_d.values()[0]) if not f=="word_index.txt"]:
        merged = [] # will contain all the words from this file, for all fields
        for field in cleaned_dirs_d.keys():
            cur_field_words = open(os.path.join(cleaned_dirs_d[field], file_id)).readline()
            cur_field_words = ["%s_%s" % (field.upper(), word) for word in cur_field_words.split(" ")]
            merged.extend(cur_field_words)
        
        # now write the merged file out
        fout = open(os.path.join(out_dir, file_id), 'w')
        fout.write(" ".join(merged))
        fout.close()


def gen_train_test(fv_file_path, n, train_path="train", test_path="test"):
    # splits the train file found @ fv_file_path
    # into train/test sets; the train set will
    # comprise the first n instances. 
    lines = open(fv_file_path).readlines()
    train = lines[:n]
    test = lines[n:]

    fout = open(train_path, 'w')
    fout.write("".join(train))
    fout.close()

    fout = open(test_path, 'w')
    fout.write("".join(test))
    fout.close()


def scl(src_folder, src_fv_name, tgt_folder, tgt_fv_name, \
            out_path, pivot_features=None, num_pivots=50, \
            src_word_list_name="word_index.txt", \
            tgt_word_list_name="word_index.txt"):
    '''
    structural correspondence learning (Blitzer, McDonal, Pereira; ACL 06)

    note: this will only work on Python 2.6+!

    pivot_features -- use this argument if you want to specify a list of
                    pivot features to use explicitly. the list should
                    be the words themselves (not indices). if it is None, the top 
                    num_pivots features (default 50, per Blitzer et al, section 7.1)
                    occuring in both the source and target domain will be used as pivots
    
    '''
    src_fv = os.path.join(src_folder, src_fv_name)
    tgt_fv = os.path.join(tgt_folder, tgt_fv_name)

    src_word_list_path = os.path.join(src_folder, src_word_list_name)
    tgt_word_list_path = os.path.join(tgt_folder, tgt_word_list_name)

    ### note that the indices_to_words are just lists!
    src_indices_to_words = eval(open(src_word_list_path).readline())
    tgt_indices_to_words = eval(open(tgt_word_list_path).readline())

    # all words, i.e., unioned lexicon
    shared_space = list(set(src_indices_to_words + tgt_indices_to_words))

    ### this changes things!! ahh
    import random
    random.shuffle(shared_space)

    shared_words_to_indices = {}
    for i, word in enumerate(shared_space):
        shared_words_to_indices[word] = i

    num_features = len(shared_space)

    source_instances = [l.replace("\n", "").split(" ") for l in open(src_fv).readlines()]
    target_instances = [l.replace("\n", "").split(" ") for l in open(tgt_fv).readlines()]

    W = [] # the columns of W will be the weight vectors corresponding to the pivots

    # going to have to create a new dataset for each pivot feature
    for pf_l in pivot_features:
        ###
        # first transform the instances to a shared space
        # the labels here denote whether or not the corresponding
        # instance contains the current pivot feature (pf_l)
        ###
        src_instances_l, src_lbls_l = _transform_instances(pf_l, \
                                            source_instances, src_indices_to_words,\
                                            shared_words_to_indices, shared_space)
         
        tgt_instances_l, tgt_lbls_l = _transform_instances(pf_l, \
                                            target_instances, tgt_indices_to_words,\
                                            shared_words_to_indices, shared_space) 
      
                                                                    
        instances_l = src_instances_l + tgt_instances_l
        lbls_l = src_lbls_l + tgt_lbls_l

        ###
        # now feed this to liblinear and get the weight
        # vector out. 
        #
        # note -- the try: block is due to Windows consistently
        # yelling about memory read errors. grr. so I'm just being
        # uber-hacky about things for now. 
        success = False
        while not success:
            try:

                m =  train(lbls_l, instances_l)
                print "pf_l: %s" % pf_l

                print lbls_l.count(1)
                #w_l = m.w[:num_features]
                #pdb.set_trace()
                w_l = m.w[:num_features]
   
                success = True
            except:
                print "fail."
                pass

                                                        
        ###
        # 1. _create_dataset_for_pf
        # 2. feed this dataset to liblinear
        # 3. get the W out
        W.append(w_l)

    ### SVD of pivot predictor matrix
    W = numpy.mat(W)
    W = W.transpose()
    U, D, Vt = numpy.linalg.svd(W, full_matrices=0)
    # here's our projector. in the paper, they
    # truncate this guy, but i'm not entirely sure why?
    # I suppose if you've a ton of pivots..?
    theta = U.transpose()

    ###
    # now we augment the feature vectors with the projections
    # note that the projection itself is done inside of the
    # _to_scl_str method
    ###

    # first the source
    out_str = []
    for x in source_instances:
        out_str.append(\
            _to_scl_str(x, theta, src_indices_to_words, shared_words_to_indices, num_features))

    for x in target_instances:
        out_str.append(\
            _to_scl_str(x, theta, tgt_indices_to_words, shared_words_to_indices, num_features))



    # finally, write out to a file
    fout = open(outpath, 'w')
    fout.write("\n".join(out_str))
    fout.close()

def _to_scl_str(x, theta, indices_to_words, shared_words_to_indices, num_features):
        
        x_str = [" ".join(x[:3])] # id level1 level2
        #x = " ".join(x[3:])
        x = x[3:]

        x_dict = {}
        x_arr = numpy.zeros(num_features)

        for f_index, value in [s.split(":") for s in x]:
            f_index = int(f_index)
            the_word = indices_to_words[f_index]
            shared_index = shared_words_to_indices[the_word]
            #x_dict[shared_index] = float(value)
            x_str.append("%s:%s" % (shared_index, value))
            x_arr[shared_index] = float(value)
        
        ###
        # project to shared space; append proction to
        # feature vector
        projection = numpy.dot(theta, x_arr).tolist()[0]
        for j in range(len(projection)):
            # append the projection to the feature vector
            x_str.append("%s:%s" % ((num_features+j), projection[j]))

        return " ".join(x_str)

def _transform_instances(pf_l, instances, indices_to_words, \
                            shared_words_to_indices, shared_space):
    instances_l = []
    lbls_l = []
    
    # just IN CASE (get it???)
    pf_l = pf_l.lower()

    total_slows = 0
    yes_slows = 0

    for x in instances:
        # for debuggign
        cur_doc = []
        cur_doc_shared = []
        slow_in_it = False

        cur_inst = {} # dictionary mapping feature indices to values
        y = -1 # assume the word pf_l is not in x
        x = x[3:] # ignore id lbl1 lbl2
        for f_index, value in [s.split(":") for s in x]:
            f_index = int(f_index)
            the_word = indices_to_words[f_index]
            cur_doc.append(the_word)
            
            if the_word.startswith("MH_"):
                # TODO remove numbers -- TEMP HACK need regex
                the_word.replace("11061", "").replace("57", "")

            ### remove field tokens
            the_word = the_word.replace("AB_", "").\
                            replace("TI_", "").replace("MH_", "")

            if the_word == pf_l or the_word + "s" == pf_l:
                # note also that in this case we *do not*
                # add the word to the instances feature
                # vector, since this would make it trivially
                # easy to classify
                y = 1
            else:
                # we retrieve the word again, since we butchered
                # it above to check it against the pivot feature
                the_word = indices_to_words[f_index]
                shared_index = shared_words_to_indices[the_word]
                cur_doc_shared.append(shared_space[shared_index])
                cur_inst[shared_index] = float(value)
        
        # feats= [int(s.split(":")[0]) for s in x]
        # words = [indices_to_words[xid] for xid in feats]

        if slow_in_it:
            total_slows += 1
            if y == 1:
                yes_slows += 1
            #pdb.set_trace()

        instances_l.append(copy.deepcopy(cur_inst))
        lbls_l.append(y)

    return (instances_l, lbls_l)



def output_word_probs(input_folder, fv_name, 
                            word_list_path, \
                            fout_path=None):
    '''
    output a file with rows as follows:

        word | index | p(y=1|word) | p(y=1|word)
    '''
    ###
    # we need access to the learners.
    ###
    sys.path.append("../modeling/curious_snake")
    import dataset
    import learners.base_nb_learner as base_nb_learner # naive bayes
    import pickle

    path_to_src = os.path.join(input_folder, fv_name)
    source_dataset = dataset.build_dataset_from_file(path_to_src)
    
    word_list = eval(open(word_list_path).readline())


    ###
    # construct a naive bayes learner over the source
    # dataset.
    nb_learner_src = base_nb_learner.BaseNBLearner([source_dataset])
    nb_learner_src.label_all_data() # legal, because this is the source data
    nb_learner_src.rebuild_models(True)

    cond_probs = nb_learner_src.models[0].conditional_probs
    out_d = {}
    for f_j in cond_probs.keys():
        word = word_list[f_j]
        out_d[word]=cond_probs[f_j]

    '''
    out_str = ["word\tsource_index\tp(y=1|x)\tp(y=-1|x)"]
    for f_j in cond_probs:
        if f_j % 100 == 0:
            print "on feature %s" % f_j

        word = source_word_list[f_j]
        if word in target_word_list:
            tgt_index = target_word_list.index(word)
            target_word_list.pop(word)
            p_1, p_0 = cond_probs[f_j][1], cond_probs[f_j][-1]
            out_str.append("%s\t%s\t%s\t%s\t%s" % (word, f_j, tgt_index, p_1, p_0))
    '''
    if fout_path is None:
        fout_path = os.path.join(input_folder, "%s_probs_d" % fv_name)

    fout = open(fout_path, 'w')
    pickle.dump(out_d, fout)
    fout.close()
    #pdb.set_trace()


def pooling_adapt(input_folder, src_fv_name, tgt_fv_name, fout_path,\
                    shared_word_list_name="shared_word_list.txt",
                    N=200):
    '''
    note that we assume instances are here represented in
    a *shared* space (comprising the words in shared_word_list_name) 

    sample call

        pooling_adapt("_transfer/prostate", 
                      "prostate_as_ab_mh_ti_shared_representation", 
                      "prostate_ab_mh_ti_shared_representation", 
                      "_transfer/prostate/pooling_adapt", 
                      "shared_word_list.txt")
    '''

    ###
    # we need access to the learners.
    ###
    sys.path.append("../modeling/curious_snake")
    import dataset
    import learners.base_nb_learner as base_nb_learner # naive bayes

    path_to_src = os.path.join(input_folder, src_fv_name)
    source_dataset = dataset.build_dataset_from_file(path_to_src)
    
    path_to_tgt = os.path.join(input_folder, tgt_fv_name)
    target_dataset = dataset.build_dataset_from_file(path_to_tgt)

    ###
    # make sure that the ids are unique
    s1 = set(source_dataset.instances.keys())
    s2 = set(target_dataset.instances.keys())
    if len(s1.intersection(s2))  > 0:
        raise Exception, "whoops, you don't have unique ids bro."
    
    # the start index for psuedo instances 
    start_index = max(max(source_dataset.instances.keys()), \
                      max(target_dataset.instances.keys())) + 1
                    
    shared_word_list = \
            eval(open(os.path.join(input_folder, shared_word_list_name)).readline())
    ###
    # construct a naive bayes learner over the source
    # dataset.
    nb_learner_src = base_nb_learner.BaseNBLearner([source_dataset])
    nb_learner_src.label_all_data() # legal, because this is the source data
    nb_learner_src.rebuild_models(True)

    ####
    # construct a learner on the *target*
    # we're going to cheat for now on the 
    # features -- this is our ORACLE
    nb_learner_test = base_nb_learner.BaseNBLearner([target_dataset])
    nb_learner_test.label_all_data() # CHEATING
    nb_learner_test.rebuild_models(True)

    best_features = _top_k_features(nb_learner_src.models[0].conditional_probs, 100)
    best_feature_words = [shared_word_list[f[0]] for f in best_features]


    # map features to their \deltas in the target task
    conditional_probs = nb_learner_test.models[0].conditional_probs
    test_features_to_deltas = {}
    for f_j in conditional_probs:
        test_features_to_deltas[f_j] = conditional_probs[f_j][1]-conditional_probs[f_j][-1]
    
    # what are the deltas in the test set for the features that were the 'best'
    # in the source set?
    best_f_deltas_in_test = [test_features_to_deltas[f_star[0]] for f_star in best_features]
    
    # generate our psuedo instance string
    psuedo_instance_str = _psuedo_positives(best_features, N, start_index)

    ###
    # assemble a string containing
    #       <the source dataset>
    #       # target dataset starts here
    #       <the target dataset>
    #       # psuedo-instances start here   
    #       <psuedo-instances>
    out_str = open(path_to_src).readlines()
    out_str.append("\n# target dataset starts here! \n")
    out_str.extend(open(path_to_tgt).readlines())
    out_str.append("\n# psuedo instances start here! \n")
    out_str.append("\n".join(psuedo_instance_str))
    fout = open(fout_path, 'w')
    pdb.set_trace()
    fout.write("".join(out_str))
    fout.close()
    print "ok -- file written to %s" % fout_path
    #pdb.set_trace()


def _psuedo_positives(feature_ids, N, start_feature_id, m=20):
    return_str = []
    for i in xrange(N):
        cur_str = ["%s 1 ? " % (start_feature_id + i)]
        feats = random.sample(feature_ids, m)
        for f_j in feats:
            cur_str.append("%s:1.0" % f_j[0])

        
        return_str.append(" ".join(cur_str))
    return return_str
        

def _top_k_features(conditional_probs, k):
    ''' top k most discriminative features '''
    features_to_deltas = {}
    ## greatest margin between estimate (P) given c=1 versus
    # for c=-1.
    for f_j in conditional_probs:
        features_to_deltas[f_j] = conditional_probs[f_j][1]-conditional_probs[f_j][-1]
    
    sorted_features = sorted(features_to_deltas.items(), key=itemgetter(1), reverse=True)
    return sorted_features[:k]


def map_src_to_tgt(src_folder, src_fv_name, tgt_folder, \
                    src_word_list_name="word_index.txt", \
                    tgt_word_list_name="word_index.txt"):

    '''
    this simple transformation simply maps the src instances
    into the target instances' feature space.
    '''
    src_word_list_path = os.path.join(src_folder, src_word_list_name)
    tgt_word_list_path = os.path.join(tgt_folder, tgt_word_list_name)

    src_fv = os.path.join(src_folder, src_fv_name)
    
    src_indices_to_words = eval(open(src_word_list_path).readline())
    tgt_indices_to_words = eval(open(tgt_word_list_path).readline())

    tgt_words_to_indices = {}
    for i, word in enumerate(tgt_indices_to_words):
        tgt_words_to_indices[word] = i

    # now we transform the target instances
    mapped_src = []
    src_vector = [l.replace("\n", "").split(" ") for l in open(src_fv).readlines()]
    for line_index, x in enumerate(src_vector):
        if line_index % 100 == 0:
            print "on line %s of %s" % (line_index, src_fv_name)
        x_transformed = x[:3] # grap id lbl1 lbl2
        total_count, intersect_count = 0, 0
        for src_index, value in [s.split(":") for s in x[3:]]:
            word = src_indices_to_words[int(src_index)]
            total_count += 1
            if word in tgt_indices_to_words:    
                intersect_count += 1
                transformed_index = tgt_words_to_indices[word]
                x_transformed.append("%s:%s" % (transformed_index, value))
        
        mapped_src.append(x_transformed)
    print "ok -- done with %s. intersction size: %s out of %s" % \
                (src_fv_name, intersect_count, total_count)
    return mapped_src

def easy_adaptation(src_folder, src_fv_name,\
                    tgt_folder, tgt_fv_name, \
                    out_path, \
                    src_word_list_name="word_index.txt", \
                    tgt_word_list_name="word_index.txt", \
                    shared_space_only=False,
                    separate_files=False, enforce_unique_ids=True):
    '''
        Daume III's frustratingly easy domain adaptation (ACL, 2007).

        the feature vectors in the respective directories will be mapped
        into a shared representation. the only difference between src and
        tgt here is that the latter will be at the end of the file (the 
        former at the start), so the first m entries of the produced feature
        vector file will correspond to the source data (assuming the source
        data comprises m instances) and the rest to target. labels will be
        kept.

        both folders are to contain feature vectors (src/tgt_fv_name) and word
        lists which are python lists (literally, strings; not pickled) like:
            [hello, world, ... , blah]
        and the ith word corresponds to index i (obviously).

        out_path is a directory

        shared_space_only will simply map both feature sets
        to a shared space (useful for, e.g., using our pooling adapt method)
    '''
    src_word_list_path = os.path.join(src_folder, src_word_list_name)
    tgt_word_list_path = os.path.join(tgt_folder, tgt_word_list_name)

    src_fv = os.path.join(src_folder, src_fv_name)
    tgt_fv = os.path.join(tgt_folder, tgt_fv_name)

    # map source/target words to thei new indices
    ### note that the indices_to_words are just lists!
    src_indices_to_words = eval(open(src_word_list_path).readline())
    tgt_indices_to_words = eval(open(tgt_word_list_path).readline())

    all_words = src_indices_to_words + tgt_indices_to_words
    print "total words in two datasets: %s" % len(all_words)
    shared_word_list = list(set(all_words))
    print "length of shared representation: %s" % len(shared_word_list)
    print "total words in common: %s" % (len(all_words) - len(shared_word_list))
 
    _mkdir(out_path)
    outf = open(os.path.join(out_path, "shared_word_list.txt"), 'w')
    outf.write(str(shared_word_list))
    outf.close()
    shared_size = len(shared_word_list)

    words_to_shared_indices = {}
    for i, word in enumerate(shared_word_list):
        words_to_shared_indices[word] = i

    transformed_representation = []
    ''' 
         map x -> <x, x, 0> in source and
             x -> <x, 0, x> in target
    '''
    source_vector = [l.replace("\n", "").split(" ") for l in open(src_fv).readlines()]
    list_of_source_xids = []
    for x in source_vector:
        x_transformed = x[:3] #  id lbl1 lbl2
        list_of_source_xids.append(int(x_transformed[0]))
        for src_index, value in [s.split(":") for s in x[3:]]:
            word = src_indices_to_words[int(src_index)]
            transformed_index = words_to_shared_indices[word]

            ### <x, x, 0>
            # *shared* version
            x_transformed.append("%s:%s" % (transformed_index, value))

            if not shared_space_only:
                # *specific* version
                x_transformed.append("%s:%s" % ((transformed_index + shared_size), value))

        transformed_representation.append(" ".join(x_transformed))
    
    if separate_files:
        # write out the src in the transformed space
        fout = open(os.path.join(out_path, "%s_shared_representation" % src_fv_name), 'w')
        fout.write("\n".join(transformed_representation))
        fout.close()
        transformed_representation = []
    else:
        transformed_representation.append("# here is the start of the target file (%s)" % tgt_fv_name)
        print "first target instance will be @ %s" % len(transformed_representation)

    # now we transform the target instances
    target_vector = [l.replace("\n", "").split(" ") for l in open(tgt_fv).readlines()]
    for x in target_vector:
        x_transformed = x[:3] # grap id lbl1 lbl2
        cur_x_id = int(x_transformed[0])
        if enforce_unique_ids and cur_x_id in list_of_source_xids:
            # we're going to append '000' to the id
            # this is kind of hacky
            x_transformed[0] = str(cur_x_id) + '000'

        for tgt_index, value in [s.split(":") for s in x[3:]]:
            word = tgt_indices_to_words[int(tgt_index)]
            transformed_index = words_to_shared_indices[word]

            ### <x, 0, x>
            # *shared* version
            x_transformed.append("%s:%s" % (transformed_index, value))

            if not shared_space_only:
                # *specific* version (note the 2x)
                x_transformed.append("%s:%s" % ((transformed_index + 2*shared_size), value))
        transformed_representation.append(" ".join(x_transformed))

    if separate_files:
        fout = open(os.path.join(out_path, "%s_shared_representation" % tgt_fv_name), 'w')
        fout.write("\n".join(transformed_representation))
        fout.close()
    else:
        fout = open(os.path.join(out_path, "%s-%s" % (src_fv_name, tgt_fv_name)), 'w')
        fout.write('\n'.join(transformed_representation))
        fout.close()


def _map_indices_to_values(l):
    indices_to_vals = {}
    for i, val in enumerate(l):
        indices_to_vals[i] = val
    return indices_to_vals

#############################
#   File encoding routines  #
#############################
def tfidf_to_file_for_lib_SVM(tfidf, pos_ids, out_path):
    out_s  = []
    for id in tfidf.keys():
        lbl = None
        if pos_ids is None:
            lbl = "?"
        else:
            lbl = -1
            if id in pos_ids:
                lbl = 1
        out_s.append(lib_svm_str(lbl, tfidf[id]))

    open(out_path, "w").write("\n".join(out_s))
          
    
def tfidf_to_file_for_lib_SVM_multi_label(tfidf, level1_pos_ids, level1_neg_ids, 
                                                 level2_pos_ids, out_path):
    ''' For the abstract screening scenario, in which there are two 'levels' of labels. '''
    out_s  = []
    
    if level1_pos_ids is not None and len(level1_pos_ids) > 0:
        if type(tfidf.keys()[0]) != type(level1_pos_ids[0]):
            print "keys and level 1 ids are not of the same type!!!"
            pdb.set_trace()
        
    for id in tfidf.keys():
        level1_lbl, level2_lbl = None, None
        
        if level1_pos_ids is None or (id not in level1_neg_ids and id not in level1_pos_ids):
            level1_lbl = "?"
        else:
            level1_lbl = -1
            if id in level1_pos_ids:
                level1_lbl = 1
                
        if level2_pos_ids is None:
            level2_lbl = "?"
        else:
            level2_lbl = -1
            
            if id in level2_pos_ids:
                level2_lbl = 1
        out_s.append(lib_svm_str_multi_label(id, level1_lbl, level2_lbl, tfidf[id]))

    open(out_path, "w").write("\n".join(out_s))
    
    
def lib_svm_str(lbl, x):
    ''' Returns a (sparse-format) feature vector string for the provided example'''
    x_str = [str(lbl)]
    for i in range(len(x)):
        if x[i] > 0.0:
            x_str.append("%s:%s" % (i, x[i]))
    return " ".join(x_str)
  
def lib_svm_str_multi_label(id, level1_lbl, level2_lbl, x):
    # 12/17/10 -- we are now assuming that x is a *dictionary*!
    # this is going to break if you try to do tf-idf instead
    # of binary
    if level2_lbl is None:
        level2_lbl = "?"
    x_str = [str(id), str(level1_lbl), str(level2_lbl)]
    if isinstance(x, dict):
        for i, val in x.items():
            x_str.append("%s:%s" % (i, val))
    else:
        for i, val in enumerate(x):
            if val > 0:
                x_str.append("%s:%s" % (i, val))
            
    return " ".join(x_str)

def generate_weka_file(labels, frequency_vectors, words, out_path):
    '''
    Builds and writes out a WEKA formatted file with the word frequencies 
    as attributes for each instance.
    '''
    weka_str = ["@RELATION abstracts"]
    for i in range(len(words)):
        # e.g.,: @ATTRIBUTE sepallength NUMERIC
        weka_str.append("@ATTRIBUTE " + words[i] + " INTEGER")
    weka_str.append("@ATTRIBUTE class {0,1}")
    weka_str.append("\n@DATA")
    for instance in range(len(frequency_vectors)):
        weka_str.append(build_weka_line_str(labels[instance], frequency_vectors[instance]))
    f_out = open(out_path, "w")
    f_out.write("\n".join(weka_str))
    
  
def build_weka_line_str(label, word_freq):
    '''
    Create a WEKA style (ARFF) line for the document associated with the provided 
    wordFreq parameter.
    '''
    line = ["{"]
    for i in range(len(wordFreq)):
        # Sparse formatting: Give the attribute 'index' first, then the value if it's non zero
        if wordFreq[i] > 0:
            line.append(str(i) + " " + str(wordFreq[i]))
    return ", ".join(line) + ", " + str(len(wordFreq)) + " " + str(label) + "}"
    
def _mkdir(newdir):
    """
    works the way a good mkdir should
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make them as well
    """
    if os.path.isdir(newdir):
        pass
    elif os.path.isfile(newdir):
        raise OSError("a file with the same name as the desired " \
                      "dir, '%s', already exists." % newdir)
    else:
        head, tail = os.path.split(newdir)
        if head and not os.path.isdir(head):
            _mkdir(head)
        if tail:
            os.mkdir(newdir)


################################################################
#
#  Unit tests! Use nose 
#           [http://somethingaboutorange.com/mrl/projects/nose/0.11.1/]. 
#  
#   e.g., while in this directory:
#           > nosetests tfidf2
#
################################################################
'''
@nose.with_setup(clean_datasets, remove_cleaned)
def binary_encode_test():
    bow = build_bag_of_words_over_dir(os.path.join("test_corpus", "cleaned"), min_word_count=1, 
                                                            binary_encode = True)
    # hand verified
    assert(bow["1"] == [0.0, 1.0, 0.0, 1.0])
    assert(bow["2"] == [1.0, 0.0, 1.0, 1.0])

    
def clean_datasets():
    clean_path = os.path.join("test_corpus", "cleaned")
    _mkdir(clean_path)
    clean_up_docs("test_corpus", out_dir=clean_path)
     
def clean_paths():
    return [os.path.join("test_corpus", "cleaned", "%s.txt") % (i+1) for i in range(2)]

def remove_cleaned():
    print clean_paths
    for f in clean_paths():
        os.remove(f)
    
@nose.with_setup(clean_datasets, remove_cleaned)
def tf_idf_test():
    bow = build_bag_of_words_over_dir(os.path.join("test_corpus", "cleaned"), min_word_count=1)
    # these are hand calculated / verified
    assert(bow["1"] == [0.0, 1.0, 0.0, 0.0])
    assert(bow["2"] == [0.70710678118654746, 0.0, 0.0, 0.70710678118654746])
   

@nose.with_setup(clean_datasets, remove_cleaned)
def clean_docs_test():
    d1, d2 = [open(p, "r").readline() for p in clean_paths()]
    assert(d1 == "humans monkeys")
    assert(d2 == "snakes like monkeys")
'''



from abstrackr.tests import TestController
from abstrackr import model

from nose.plugins.attrib import attr  # Decorator to mark tests
                                      # Use "nosetests -a author=jj" to
                                      # run only marked tests or
                                      # "nosetests -a '!author=jj" to
                                      # exclude marked tests

import abstrackr.controllers.controller_globals as cg

Session = model.Session


class TestControllerGlobals(TestController):

    @attr(author='jj', controller='controller_globas')
    def test_get_project_member_ids(self):
        """ Return list of member ids

        Method takes a project id and returns should returns a
        list of user ids participating in the project.

        """

        u1 = model.User()
        u2 = model.User()
        u3 = model.User()
        p1 = model.Project()

        p1.members.extend([u1, u2, u3])
        Session.add(u1)
        Session.add(u2)
        Session.add(u3)
        Session.add(p1)
        Session.commit()

        member_id_list = cg._get_project_member_ids(p1.id)
        assert member_id_list == [u1.id, u2.id, u3.id]

        # Clean up
        Session.delete(u1)
        Session.delete(u2)
        Session.delete(u3)
        Session.delete(p1)
        Session.commit()



from abstrackr.lib.base import BaseController, render

from repoze.what.plugins.pylonshq import ActionProtector
from repoze.what.predicates import not_anonymous

from pylons import tmpl_context as c, url


class StaticPagesController(BaseController):

    @ActionProtector(not_anonymous())
    def help(self):
        c.root_path = url('/', qualified=True)
        return render("/static_pages/help.mako")

    @ActionProtector(not_anonymous())
    def citing(self):
        c.root_path = url('/', qualified=True)
        return render("/static_pages/citing.mako")



#!/usr/bin/env python
# encoding: utf-8
#
# 
# The MIT License
# 
# Copyright (c) 2009 Byron C Wallace
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
Byron C Wallace
Tufts Medical Center
pubmedpy.py
--

Example use:

>python pubmed_fetchr.py -e bwallace@tuftsmedicalcenter.org -s biopython

(NCBI wants your email address for the web services stuff). -s is the search string, here we search 
for abstracts related to "biopython". 

"""

# std libraries
import sys
import getopt
import pdb
import os
from optparse import OptionParser
import Bio
from Bio import Entrez
from Bio import Medline
import EUtils
from EUtils import ThinClient

# home-grown
import tfidf2 

def main(argv=None):
  parser = OptionParser()
  parser.add_option("-e", "--email", action="store", type="string", dest="email")
  parser.add_option("-s", "--search_string", action="store", type="string", dest="search_str")
  (options, args) = parser.parse_args()
  
  Entrez.email = options.email
  records = article_search("pubmed", options.search_str)
  print "Writing out title and abstract data..."
  write_out_fields(["TI", "AB"], records)
  print("fin.")
    
def set_email(email):
    Entrez.email = email

def fetch_and_encode(article_ids, out_dir, binary_features=False, 
                                    labels=None, fields = ["AB", "TI"], out_f_name = ""):
    '''
    First fetches from the web, then encodes them.
    '''
    # first, fetch the articles
    fetch_and_write_out(article_ids, out_dir, fields = fields)
    
    for field in fields:   
        print "encoding %s..." %field
        # now, clean and encode them
        out_for_field = os.path.join(out_dir, field)
        tfidf2.encode_docs(out_for_field, os.path.join(out_for_field, "encoded"), out_f_name + field, lbl_dict = labels)
    print "finito."
    

def fetch_articles(article_ids):
    print "fetching abstracts..."
    print article_ids
    #handle = Entrez.efetch(db="pubmed",id=article_ids,rettype="medline",retmode="text")
    
    dbids = EUtils.DBIds("pubmed", [str(xid) for xid in article_ids])
    client = ThinClient.ThinClient()

    # fix for new NCBI API
    handle = client.efetch_using_dbids(dbids=dbids, rettype='medline', retmode='text')
    #handle = EUtils.efetch(db="pubmed",id=article_ids,rettype="medline",retmode="text")
    records = Medline.parse(handle)
    print "Done." 
    return records   
    
def batch_fetch(article_ids, batch_size=20):
    all_records = []
    total = len(article_ids)
    fetched_so_far = 0
    while fetched_so_far < total:
      records = fetch_articles(article_ids[fetched_so_far:fetched_so_far+batch_size])
      fetched_so_far += batch_size
      all_records.extend([r for r in records])
      print "fetched so far, total to fetch: %s, %s" % (fetched_so_far, total)
  
    return all_records
  

  
def article_search(db, search_str):
    handle = Entrez.esearch(db=db,term=search_str,retmax=1000)
    record = Entrez.read(handle)
    article_ids = record["IdList"]
    print "Found %s articles" % len(article_ids)
    return batch_fetch(article_ids)
  
 
def write_to_tsv(field_keys, records, fpath):
    header_line = "\t".join(field_keys)
    all_lines = [header_line]
    for r in records:
        cur_line = []
        for field in field_keys:
            cur_line.append(str(r[field]))
        all_lines.append("\t".join(cur_line))
    outf = open(fpath, 'w')
    outf.write("\n".join(all_lines))
    
def write_out_fields(field_keys, records, base_out_dir):
    # make a directory for each field
    if not os.path.isdir(base_out_dir):
        os.mkdir(base_out_dir)
        
    for field in field_keys:
      field_dir = os.path.join(base_out_dir, field)
      if not os.path.isdir(field_dir):
        os.mkdir(field_dir)
    # write out 
    for record in records:
      for field in field_keys:
        out_f = open(os.path.join(base_out_dir, field, record["PMID"]), "w")
        cur_field = None
        try:
          cur_field = record[field]
        except Exception, inst:
          cur_field = ""
          
        if isinstance(cur_field, list):
            cur_field = ", ".join(cur_field)
        out_f.write(cur_field)
        out_f.close()




if __name__ == "__main__":
  sys.exit(main())



import pdb 

import numpy as np

###
# feature extraction & encoding
import sklearn
from sklearn.feature_extraction.text import TfidfVectorizer

class Dataset:

    def __init__(self, ids, titles, abstracts, mesh, lbl_dict, name=None, 
                    stop_word_fpath="/Users/bwallace/dev/abstrackr-web/abstrackr/lib/stop_list.txt"):
        '''
        assumes the ordering of ids is the same as the ordering of titles 
        and abstracts! 
        '''
        self.all_ids = ids
        self.N = len(self.all_ids)
        self.ids_to_indices = dict(zip(self.all_ids, range(self.N)))

        for i, id_ in enumerate(self.all_ids):
            self.ids_to_indices[id_] = i 
        
        self.titles = self._replace_None(titles)
        self.abstracts = self._replace_None(abstracts)
        self.mesh = self._replace_None(mesh)
        
        assert(len(ids) == len(titles) == len(abstracts))
        
        self.lbl_dict = lbl_dict

        print "done. now reading labels in..."
        self._setup_lbl_vecs()
        self.stop_word_list_path = stop_word_fpath
        #self.stop_word_list_path = "/Users/abstrackr-user/Hive/abstrackr/abstrackr/lib/stop_list.txt"
        self._load_stopwords()
        print "alright -- encoding!"
        self.encode()

    def _replace_None(self, x):
        for i, x_i in enumerate(x):
            if x_i is None:
                x[i] = ""
        return x

    def __len__(self):
        return self.N

    def get_train_X_y(self):
        train_indices = self._get_train_indices()
        if len(train_indices) == 0:
            raise Exception, "nothing has been labeled yet!"
        
        return self.get_X_y(indices=train_indices)

    def _get_train_indices(self):
        return [self.ids_to_indices[id_] for id_ in self.labeled_ids]

    def _get_test_indices(self):
        return [self.ids_to_indices[id_] for id_ in self.unlabeled_ids]

    def get_test_X_y(self):
        test_indices = self._get_test_indices()
        return self.get_X_y(indices=test_indices)

    def get_test_ids(self):
        return self.unlabeled_ids

    def get_train_level_1_y(self):
        train_indices = self._get_train_indices()
        return self.l1s[train_indices].tolist()

    def get_X_y(self, indices=None):
        ids = []
        if indices is not None:
            if len(indices) == 0:
                raise Exception, "list of indices is empty."
            return self.abstracts_X[indices], self.titles_X[indices], self.mesh_X[indices], self.l1s[indices]
        return self.abstracts_X, self.titles_X, self.mesh_X, self.l1s

    def encode(self):
        self.abstracts_vectorizer = TfidfVectorizer(ngram_range=(1,2), max_features=50000, min_df=3, 
                                                    stop_words=self.stopwords)
        print "vectorizing abstracts..."
        self.abstracts_X = self.abstracts_vectorizer.fit_transform(self.abstracts)
        print "done. %s abstract features. now titles ..." % self.abstracts_X.shape[1]

        self.titles_vectorizer = TfidfVectorizer(ngram_range=(1,2), max_features=50000, min_df=3, 
                                                    stop_words=self.stopwords)
        self.titles_X = self.titles_vectorizer.fit_transform(self.titles)
        print "ok. %s title features." % self.titles_X.shape[1]
        self.mesh_X = None
        
        print "and finally, mesh..."
        self.mesh_vectorizer = TfidfVectorizer(max_features=50000, min_df=3)
        self.mesh_X = self.mesh_vectorizer.fit_transform(self.mesh)
        print "ok -- %s mesh features." % self.mesh_X.shape[1]

        print "all encoded!"

    def _load_stopwords(self):
        self.stopwords = [w.strip() for w in open(self.stop_word_list_path, 'rU').readlines()]

    def _setup_lbl_vecs(self):
        self.l1s = np.zeros(self.N)
        self.labeled_count = 0
        self.labeled_ids, self.unlabeled_ids = [], []
        for i, id_ in enumerate(self.all_ids):
            # now set the labels
            if id_ not in self.lbl_dict:
                # not yet labeled
                self.l1s[i] = None
                self.unlabeled_ids.append(id_)
            else:
                self.labeled_ids.append(id_)
                self.labeled_count += 1
                if self.lbl_dict[id_] > 0:
                    self.l1s[i] = 1.0
                else:
                    self.l1s[i] = -1.0

        count_ones = lambda a: len([x for x in a if x == 1])
        print "%s total labeled; %s includes" % (
                        self.labeled_count,
                        count_ones(self.l1s))


from abstrackr.tests import TestController

from abstrackr import model

Session = model.meta.Session

class TestReviewModel(TestController):
    """ Make sure db tables have the correct columns """


    def test_project(self):
        assert hasattr(model.Project, 'id')
        assert hasattr(model.Project, 'leader_id')
        assert hasattr(model.Project, 'initial_assignment_id')
        assert hasattr(model.Project, 'name')
        assert hasattr(model.Project, 'description')
        assert hasattr(model.Project, 'code')
        assert hasattr(model.Project, 'screening_mode')
        assert hasattr(model.Project, 'tag_privacy')
        assert hasattr(model.Project, 'num_labels_thus_far')
        assert hasattr(model.Project, 'sort_by')
        assert hasattr(model.Project, 'initial_round_size')
        assert hasattr(model.Project, 'date_created')
        assert hasattr(model.Project, 'date_modified')

        # This is the backref from the "user" table
        assert hasattr(model.Project, 'leader')
        assert hasattr(model.Project, 'members')

    def test_citation(self):
        assert hasattr(model.Citation, 'id')
        assert hasattr(model.Citation, 'project_id')
        assert hasattr(model.Citation, 'pmid')
        assert hasattr(model.Citation, 'refman')
        assert hasattr(model.Citation, 'title')
        assert hasattr(model.Citation, 'abstract')
        assert hasattr(model.Citation, 'authors')
        assert hasattr(model.Citation, 'journal')
        assert hasattr(model.Citation, 'publication_date')
        assert hasattr(model.Citation, 'keywords')
        assert hasattr(model.Citation, 'tasks')

    def test_priority(self):
        assert hasattr(model.Priority, 'id')
        assert hasattr(model.Priority, 'project_id')
        assert hasattr(model.Priority, 'citation_id')
        assert hasattr(model.Priority, 'priority')
        assert hasattr(model.Priority, 'num_times_labeled')
        assert hasattr(model.Priority, 'is_out')
        assert hasattr(model.Priority, 'locked_by')
        assert hasattr(model.Priority, 'time_requested')

    def test_tagtypes(self):
        pass

    def test_tags(self):
        pass

    def test_note(self):
        pass

    def test_labeledfeature(self):
        pass

    def test_label(self):
        pass

    def test_users_projects(self):
        # Create objects to test relationship (many-to-many)
        u1 = model.User()
        u2 = model.User()
        p1 = model.Project()
        p1.members.append(u1)
        p1.members.append(u2)
        Session.add(p1)
        Session.commit()

        # Get user objects now that they should have associated projects
        u1_ = Session.query(model.User).filter_by(id=u1.id).one()
        u2_ = Session.query(model.User).filter_by(id=u2.id).one()

        # Test that relationship to project exists
        assert u1_.projects == [p1]
        assert u2_.projects == [p1]

        # Now delete the project
        Session.delete(p1)
        Session.commit()

        # Verify that relationship no longer exists
        # (ie. users_projects_table entry has been removed)
        assert u1_.projects == []
        assert u2_.projects == []

        # Let's clean up
        Session.delete(u1)
        Session.delete(u2)
        Session.commit()

    def test_assignment(self):
        pass

    def test_task(self):
        # This is the backref from the "Citations" table
        assert hasattr(model.Task, 'citations')

    def test_fixedtask(self):
        pass

    def test_encodestatus(self):
        pass

    def test_predictionsstatus(self):
        pass

    def test_prediction(self):
        pass

    def test_resetpassword(self):
        pass

    def test_group(self):
        pass

    def test_permission(self):
        pass

    def test_user(self):
        assert hasattr(model.Priority, 'id')

    def test_grouppermission(self):
        pass

    def test_usergroup(self):
        pass



import logging

from pylons import config, request, response, session, tmpl_context as c, url
from pylons.controllers.util import abort, redirect
from pylons.controllers.util import redirect

import abstrackr.model as model
from abstrackr.model.meta import Session
from abstrackr.lib.base import BaseController, render
import abstrackr.controllers.controller_globals as controller_globals # shared querying routines

from repoze.what.predicates import not_anonymous, has_permission
from repoze.what.plugins.pylonshq import ActionProtector

import turbomail
import smtplib
import string
import random

from sqlalchemy import func

# validation stuff
from pylons.decorators import validate
import abstrackr.model.form as form

log = logging.getLogger(__name__)

class AccountController(BaseController):

    def login(self):
        """
        This is where the login form should be rendered.
        Without the login counter, we won't be able to tell if the user has
        tried to log in with wrong credentials.
        """
        identity = request.environ.get('repoze.who.identity')
        came_from = str(request.GET.get('came_from', '')) or \
                     url(controller='account', action='welcome')
        if identity:
            session['flash'] = 'Login successful.'
            session.save()
            redirect(url(came_from))
        else:
            c.came_from = came_from
            c.login_counter = request.environ['repoze.who.logins'] + 1
            return render('/accounts/login.mako')

    def email_test(self):
        from turbomail import Message
        msg = Message('byron.wallace@gmail.com', 'byron.wallace@gmail.com', 'Subject')
        msg.plain = "Foo Bar"
        try:
            msg.send()   # Message will be sent through the configured manager/transport.
        except Exception, err:
            print err

    def recover_password(self):
        c.pwd_msg = ""
        return render('/accounts/recover.mako')

    def reset_password(self):
        user_email = request.params['email']
        user_for_email = controller_globals._get_user_from_email(user_email)
        if user_for_email:
            token = self.gen_token_to_reset_pwd(user_for_email)
            message = """
                        Hi, %s. \n
                        Someone (hopefully you!) asked to reset your abstrackr password.
                        To do so, follow this link:\n
                        \t %saccount/confirm_password_reset/%s.\n
                        Note that your password will be temporarily changed if you follow this link!
                        If you didn't request to reset your password, just ignore this email.
                      """ % (user_for_email.fullname, url('/', qualified=True), token)
            print token
            self.send_email_to_user(user_for_email, "resetting your abstrackr password", message)
            c.pwd_msg = "OK -- check your email (and your spam folder!)"
            return render('/accounts/recover.mako')
        else:
            c.pwd_msg = """
                Well, this is awkward.
                We don't have a user in our database with email: %s.
                Try again?""" % user_email
            return render('/accounts/recover.mako')


    def confirm_password_reset(self, id):
        token = str(id)
        reset_pwd_q = Session.query(model.ResetPassword)
        # we pull all in case they've tried to reset their pwd a few times
        # by the way, these should time-expire...
        matches = reset_pwd_q.filter(model.ResetPassword.token == token).all()
        if len(matches) == 0:
            return """
                Hrmm... It looks like you're trying to reset your password, but I can't match the provided token.
                Please go back to the email that was sent to you and make sure you've copied the URL correctly.
                """
        user = controller_globals._get_user_from_email(matches[0].user_email)
        for match in matches:
            Session.delete(match)
        user._set_password(token)
        Session.commit()
        return '''
          ok! your password has been set to %s (you can change it once you've logged in).\n
          <a href="%s">log in here</a>.''' % (token, url('/', qualified=True))

    def gen_token_to_reset_pwd(self, user):
        # generate a random token for the user to reset their password; stick
        # it in the database
        make_token = lambda N: ''.join(random.choice(string.ascii_uppercase + string.digits) for x in range(N))
        reset_pwd_q = Session.query(model.ResetPassword)
        existing_tokens = [entry.token for entry in reset_pwd_q.all()]
        token_length=10
        cur_token = make_token(token_length)
        while cur_token in existing_tokens:
            cur_token = make_code(token_length)

        reset = model.ResetPassword()
        reset.token = cur_token
        reset.user_email = user.email
        Session.add(reset)
        Session.commit()
        return cur_token

    def send_email_to_user(self, user, subject, message):
        host = config['smtp_host']
        port = config['smtp_port']
        sender = config['sender']
        server = smtplib.SMTP(host=host, port=port)
        to = user.email
        body = string.join((
            "From: %s" % sender,
            "To: %s" % to,
            "Subject: %s" % subject,
            "",
            message
            ), "\r\n")

        server.sendmail(sender, [to], body)


    def my_account(self):
        c.current_user = request.environ.get('repoze.who.identity')['user']
        c.account_msg = ""
        c.account_msg_citation_settings = ""
        return render("/accounts/account.mako")

    def change_password(self):
        current_user = request.environ.get('repoze.who.identity')['user']
        if request.params["password"] == request.params["password_confirm"]:
            current_user._set_password(request.params['password'])
            Session.commit()
            c.account_msg = "ok, your password has been changed."
        else:
            c.account_msg = "whoops -- the passwords didn't match! try again."

        c.account_msg_citation_settings = ""

        return render("/accounts/account.mako")


    def create_account(self):
        if 'then_join' in request.params:
            c.then_join = request.params['then_join']

        return render('/accounts/register.mako')

    @validate(schema=form.RegisterForm(), form='create_account')
    def create_account_handler(self):
        '''
        Note that the verification goes on in model/form.py.
        '''
        # create the new user; post to db via sqlalchemy
        new_user = model.User()
        new_user.username = request.params['username']
        new_user.fullname = " ".join([request.params['first_name'], request.params['last_name']])
        new_user.experience = request.params['experience']
        new_user._set_password(request.params['password'])
        new_user.email = request.params['email']

        # These are for citation settings,
        # initialized to True to make everything in the citation visible by default.
        new_user.show_journal = True
        new_user.show_authors = True
        new_user.show_keywords = True

        Session.add(new_user)
        Session.commit()

        # send out an email
        greeting_message = """
            Hi, %s.\n

            Thanks for signing up at abstrackr (%s). You
            should be able to sign up now with username %s (only you know your password).

            This is just a welcome email to say hello, and that we've got your email.
            Should you ever need to reset your password, we'll send you instructions
            to this email. In the meantime, happy screening!

            -- The Brown EPC.
        """ % (new_user.fullname, url('/', qualified=True), new_user.username)

        try:
            self.send_email_to_user(new_user, "welcome to abstrackr", greeting_message)
        except:
            # this almost certainly means we're on our Windows dev box :)
            pass

        ###
        # log this user in programmatically (issue #28)
        rememberer = request.environ['repoze.who.plugins']['cookie']
        identity = {'repoze.who.userid': new_user.username}
        response.headerlist = response.headerlist + \
            rememberer.remember(request.environ, identity)
        rememberer.remember(request.environ, identity)


        # if they were originally trying to join a review prior to
        # registering, then join them now. (issue #8).
        if 'then_join' in request.params and request.params['then_join'] != '':
            redirect(url(controller="review", action="join", review_code=request.params['then_join']))
        else:
            redirect(url(controller="account", action="login"))



    '''
    The following methods are protected, i.e., the user must be logged in.
    '''
    @ActionProtector(not_anonymous())
    def welcome(self):
        redirect(url(controller="account", action="my_work"))

    @ActionProtector(not_anonymous())
    def back_to_projects(self):
        redirect(url(controller="account", action="my_projects"))


    @ActionProtector(not_anonymous())
    def my_work(self):

        person = request.environ.get('repoze.who.identity')['user']
        c.person = person
        user = controller_globals._get_user_from_email(c.person.email)
        if not user:
            log.error('''\
Hum...fetching user from the database returned False.
We need to investigate. Go remove the catch all in
controller_globals.py, method _get_user_from_email()
to see which OperationalError is being raised ''')

        # If somehow the user's citation settings variables don't get initialized yet,
        # then the following 3 if-else blocks should take care of it in order to avoid
        # any errors due to the values of the variables being null:
        c.show_journal = user.show_journal if not user.show_journal is None else True

        if (user.show_authors==True or user.show_authors==False):
            c.show_authors = user.show_authors
        else:
            user.show_authors = True

        if (user.show_keywords==True or user.show_keywords==False):
            c.show_keywords = user.show_keywords
        else:
            user.show_keywords = True


        # pull all assignments for this person
        assignment_q = Session.query(model.Assignment)
        all_assignments = assignment_q.filter(model.Assignment.user_id == person.id).all()
        self._set_assignment_done_status(all_assignments)
        self._clear_this_user_locks(all_assignments)

        # Build assignment completion status dictionary
        c.d_completion_status = self._get_assignment_completion_status(all_assignments)

        c.outstanding_assignments = [a for a in all_assignments if not a.done]
        # if there's an initial assignment, we'll only show that.
        assignment_types = [assignment.assignment_type for assignment in \
                                                    c.outstanding_assignments]

        #####
        # for any review that has an initial assignment, we will show
        # *only* that assignment, thereby forcining participants to
        # finish initial assignments before moving on to other
        # assignments. fix for issue #5.
        ####
        # which reviews have (outstanding) initial assigments?
        reviews_with_initial_assignments = []
        for assignment in c.outstanding_assignments:
            if assignment.assignment_type == "initial":
                reviews_with_initial_assignments.append(assignment.project_id)


        # now remove other (non-initial) assignments for reviews
        # that have an initial assignment
        filtered_assignments = [assignment for assignment in c.outstanding_assignments if \
                                assignment.project_id not in reviews_with_initial_assignments or \
                                assignment.assignment_type == "initial"]

        c.outstanding_assignments = filtered_assignments

        c.finished_assignments = [a for a in all_assignments if a.done]


        project_q = Session.query(model.Project)
        c.participating_projects = user.member_of_projects
        c.review_ids_to_names_d = self._get_review_ids_to_names_d(c.participating_projects)

        c.my_work = True
        c.my_projects = False
        return render('/accounts/dashboard.mako')

    @ActionProtector(not_anonymous())
    def _set_assignment_done_status(self, all_assignments):
        for a in all_assignments:
            b_assignment_done = controller_globals._check_assignment_done(a)
            a.done = b_assignment_done
            Session.add(a)
        Session.commit()

    @ActionProtector(not_anonymous())
    def _clear_this_user_locks(self, all_assignments):
        priorities_locked_by_this_user = self._get_all_priorities_locked_by_this_user(all_assignments)
        for p in priorities_locked_by_this_user:
            p.is_out = 0
            p.locked_by = None
            Session.add(p)
        Session.commit()

    @ActionProtector(not_anonymous())
    def _get_all_priorities_locked_by_this_user(self, all_assignments):
        locked_priorities = []
        for a in all_assignments:
            user_id = a.user_id
            project_id = a.project_id
            priorities_q = Session.query(model.Priority).filter_by(project_id=project_id).\
                                                         filter_by(locked_by=user_id)
            priorities = priorities_q.all()
            locked_priorities.extend([p for p in priorities])
        return list(set(locked_priorities))

    @ActionProtector(not_anonymous())
    def show_merge_review_screen(self):
        person = request.environ.get('repoze.who.identity')['user']
        c.person = person
        user = controller_globals._get_user_from_email(person.email)

        projects_person_leads = self._get_projects_person_leads(user)
        c.reviews = projects_person_leads

        return render('/reviews/merge_reviews.mako')


    @ActionProtector(not_anonymous())
    def my_projects(self):
        person = request.environ.get('repoze.who.identity')['user']
        c.person = person

        # Get user object from db.
        user = controller_globals._get_user_from_email(person.email)

        # Set user's show preference defaults in case they weren't set.
        if (user.show_journal==True or user.show_journal==False):
            c.show_journal = user.show_journal
        else:
            user.show_journal = True

        if (user.show_authors==True or user.show_authors==False):
            c.show_authors = user.show_authors
        else:
            user.show_authors = True

        if (user.show_keywords==True or user.show_keywords==False):
            c.show_keywords = user.show_keywords
        else:
            user.show_keywords = True

        project_q = Session.query(model.Project)
        c.leading_projects = user.leader_of_projects
        leading_project_ids = [proj.id for proj in c.leading_projects]

        c.participating_projects = [p for p in user.member_of_projects if p.id not in leading_project_ids]

        c.review_ids_to_names_d = self._get_review_ids_to_names_d(c.participating_projects)

        statuses_q = Session.query(model.PredictionsStatus)
        c.statuses = {}

        c.do_we_have_a_maybe = {}
        for project_id in leading_project_ids:

            predictions_for_review = statuses_q.filter(model.PredictionsStatus.project_id==project_id).all()
            if len(predictions_for_review) > 0 and predictions_for_review[0].predictions_exist:
                c.statuses[project_id] = True
            else:
                c.statuses[project_id] = False

            c.do_we_have_a_maybe[project_id] = False

        # Flag projects that have locked priorities
        c.projects_w_locked_priorities = self._get_projects_w_locked_priorities(leading_project_ids)

        c.my_work = False
        c.my_projects = True

        return render('/accounts/dashboard.mako')

    def _get_projects_w_locked_priorities(self, lof_project_ids):
        """Checks if projects in lof_project_ids have any locked priorities

           (listof ProjectID) -> Dictionary
        """

        dict_proj_w_locked_priorities = {}

        for i in lof_project_ids:
            has_locked_priorities = self._project_has_locked_priorities(i)
            dict_proj_w_locked_priorities[i] = has_locked_priorities

        return dict_proj_w_locked_priorities
           
    def _project_has_locked_priorities(self, project_id):
        """Returns True if project has any locked priorities, else False

           Integer -> Boolean
        """

        priorities_q = Session.query(model.Priority).\
                               filter_by(project_id=project_id).\
                               filter_by(is_out=1)
        priority = priorities_q.first()
        if priority:
            return True
        else:
            return False

    def _get_projects_person_leads(self, person):
        leading_projects = person.leader_of_projects
        return leading_projects

    def _get_review_ids_to_names_d(self, reviews):
        # make life easier on client side
        review_ids_to_names_d = {}
        for review in reviews:
            review_ids_to_names_d[review.id] = review.name
        return review_ids_to_names_d


    @ActionProtector(not_anonymous())
    def test_user_access(self):
        return 'You are inside user section'

    @ActionProtector(has_permission('admin'))
    def test_admin_access(self):
        return 'You are inside admin section'


    # This is what gets executed when the user attempts to customize the 'citation settings'.
    @ActionProtector(not_anonymous())
    def customize_citations(self):

        # Obtain the parameters.
        show_journal_str = request.params['toggle_journal']
        show_authors_str = request.params['toggle_authors']
        show_keywords_str = request.params['toggle_keywords']

        # Obtain the User object (as opposed to the auth.User object).
        cur_user = controller_globals._get_user_from_email(request.environ.get('repoze.who.identity')['user'].email)

        # Make changes to the booleans of the user object.
        cur_user.show_journal = {"Show":True, "Hide":False}[show_journal_str]
        cur_user.show_authors = {"Show":True, "Hide":False}[show_authors_str]
        cur_user.show_keywords = {"Show":True, "Hide":False}[show_keywords_str]

        # Add the changes to the database.
        Session.commit()

        # These messages appear in their designated separate locations,
        #   i.e. on the top left corner of the div that corresponds to this function.
        # This is more appropriate than having a general message on some part of the screen.
        c.account_msg = ""
        c.account_msg_citation_settings = "Citation Settings changes have been applied."

        return render("/accounts/account.mako")

    def _get_assignment_completion_status(self, assignments):
        """Returns how many labels have been recorded for each assignment ID.

        The returned object is a dictionary with the assignment ID as key
        and the number of citations screened for this assignment as the
        value

        """

        status_summary = {}

        for a in assignments:
            project_id = a.project_id
            user_id = a.user_id
            lof_labels_for_assignment = self._get_users_labels_for_assignment(project_id,
                                                                              user_id,
                                                                              a.id)

            status_summary[a.id] = len(lof_labels_for_assignment)
        return status_summary

    def _get_users_labels_for_assignment(self, project_id, user_id, assignment_id):
        """Returns a user's list of labels across a specific project"""
        labels_q = Session.query(model.Label)
        labels = labels_q.filter_by(project_id=project_id,
                                    user_id=user_id,
                                    assignment_id=assignment_id).all()
        return labels




import os, pdb, pickle, random
from operator import itemgetter
import datetime

import sqlalchemy
from sqlalchemy import *
from sqlalchemy.sql import select
from sqlalchemy.sql import and_, or_

from make_predictions import make_predictions
import tfidf2

engine = create_engine("mysql://abstrackr-user:5xl.z=Uy6d@127.0.0.1:3306/abstrackrDBP01?unix_socket=/var/mysql/mysql.sock")
metadata = MetaData(bind=engine)

####
# bind the tables
citations = Table("Citations", metadata, autoload=True)
labels = Table("Labels", metadata, autoload=True)
reviews = Table("Projects", metadata, autoload=True)
users = Table("user", metadata, autoload=True)
labeled_features = Table("labeledfeatures", metadata, autoload=True)
encoded_status = Table("encodedstatuses", metadata, autoload=True)
prediction_status = Table("predictionstatuses", metadata, autoload=True)
predictions = Table("predictions", metadata, autoload=True)
priorities = Table("priorities", metadata, autoload=True)

def get_labels_from_names(review_names):
    r_ids = get_ids_from_names(review_names)
    if len(r_ids) == 0:
        pdb.set_trace()
    return labels_for_review_ids(r_ids)

def get_ids_from_names(review_names):
    s = reviews.select(reviews.c.name.in_(review_names))
    return [review.id for review in s.execute()]

def labels_for_review_ids(review_ids):
    # this selects labels and information about the corresponding
    # labelers.
    s = select([labels, users, citations.c.abstract], \
                        and_(labels.c.reviewer_id == users.c.id,
                             labels.c.project_id.in_(review_ids),
                             citations.c.citation_id == labels.c.study_id),
                        use_labels=True)

    # unroll the result
    return [x for x in s.execute()]


def write_out_labels_for_reviews(review_sets, out_path):  
    # review_sets is a list of lists; each list therein comprises
    # n reviews corresponding to one single review
    # e.g,. [[my_review_1, my_review_2], [a diff review A, a diff review B]]
    abstract_length = lambda x: 0 if x is None else len(x.split(" "))
    all_results = []
    for reviews in review_sets:
        # first fetch the results
        result = get_labels_from_names(reviews)  
        all_results.append(result)

    # first write out the headers
    out_str = ["\t".join(["meta_review_id", "comprising"]+\
                    all_results[0][0].keys()[:-1] + ["abstract_length"])]

    for i, names in enumerate(review_sets):
        for result in all_results[i]:
            cur_str = "\t".join(["%s" % i, "%s" % "-".join(names)]+\
                                [str(x) for x in result.values()[:-1]]+\
                                [str(abstract_length(result['citations_abstract']))])
            out_str.append(cur_str)
    
    # ok, let's dump the result to file
    out_stream = open(out_path, 'w')
    out_stream.write("\n".join(out_str))
    out_stream.close()

def to_disk(base_dir, review_names=None, review_ids=None, fields=None):
    ''' 
    writes the citations (titles, abstracts, etc) and labels
    (both feature and instance) over the given reviews to disk; 
    base_dir tells the routine where to save the files.
    '''
    if not (review_ids or review_names):
        raise Exception, "you need to provide either the names or the ids of the reviews you want, dummy."
    elif review_ids is None:
        # get the ids from the names
        review_ids = get_ids_from_names(review_names)

    for review_id in review_ids:
        citations_to_disk(review_id, base_dir, fields=fields)
    
    lbls_to_disk(review_ids, base_dir)


def citations_to_disk(review_id, base_dir, fields=None):
    fields = fields or ["title", "abstract", "keywords", "authors", "journal"]
    none_to_text= lambda x: "none" if x is None else x

    s = citations.select(citations.c.project_id==review_id)
    citations_for_review = [x for x in s.execute()]

    if not os.path.exists(base_dir):
        os.mkdir(base_dir)

    for field in fields:
        print "(citations_to_disk) on field: %s" % field
        field_path = os.path.join(base_dir, field)
        if not os.path.exists(field_path):
            os.mkdir(field_path)

    for citation in citations_for_review:
        citation_id = citation['id']
        
        for field in fields:
            fout = open(os.path.join(base_dir, field, "%s" % citation_id), 'w')
            fout.write(none_to_text(citation[field]))
            fout.close()


def get_lbl_d_for_review(review_id):
    lbl_d = {}
    s = labels.select(labels.c.project_id==review_id)
    for lbl in s.execute():
        lbl_d[lbl["study_id"]]=lbl["label"]  
    return lbl_d

def lbls_to_disk(review_ids, base_dir):
    lbl_d = {}
    s = labels.select(labels.c.project_id.in_(review_ids))
    for lbl in s.execute():
        lbl_d[lbl["study_id"]]=lbl["label"]
    
    fout = open(os.path.join(base_dir, "labels.pickle"), 'w')
    pickle.dump(lbl_d, fout)                     
    fout.close()
    
    # also get labeled features
    lbl_feature_d = {}

    s =  labeled_features.select(labeled_features.c.project_id.in_(review_ids))   
    for lbld_feature in s.execute():
        lbl_feature_d[lbld_feature["term"]] = lbld_feature["label"]
    
    fout = open(os.path.join(base_dir, "labeled_features_d.pickle"), 'w')
    pickle.dump(lbl_feature_d, fout)                     
    fout.close()

def encode_review(review_id, base_dir="/Users/abstrackr-user/Hive/abstrackr/abstrackr/lib/curious_snake/data"):
    fields=["title", "abstract", "keywords"]
    
    base_dir = os.path.join(base_dir, str(review_id))
    
    # write the abstracts to disk
    to_disk(base_dir, review_ids=[review_id], fields=fields)

    # now encode them
    lbl_d = pickle.load(open(os.path.join(base_dir, "labels.pickle")))

    # we encode the three main fields in separate spaces (multi-view)
    for field in fields:
        print "\n\n\n(write_review_to_disk) on field: %s..." % field
        
        dir_path = os.path.join(base_dir, field)
        out_path = os.path.join(dir_path, "encoded")
        out_f_name = "%s_encoded" % field
        tfidf2.encode_docs(dir_path, out_path, out_f_name, \
                    lbl_dict=lbl_d, clean_first=True, binary=True, \
                    min_word_count=3, bi_grams_too=True, delete_files_after=True)

    # fetch the encoded status entry for this review, 
    # ***which we assume exists!***
    return base_dir

def update_labels(review_id, base_dir="/Users/abstrackr-user/Hive/abstrackr/abstrackr/lib/curious_snake/data"):
    new_lbls = get_lbl_d_for_review(review_id)
    fields = ["title", "abstract", "keywords"]
    for field in fields:
        # relying on conventions here
        dir_path = os.path.join(base_dir, str(review_id), field)
        f_path = os.path.join(dir_path, "encoded")
        out_f_name = "%s_encoded" % field
        f_path = os.path.join(f_path, out_f_name)
        print "updating file at %s" % f_path
        # just overwrite the old file
        tfidf2.update_labels(f_path, f_path, new_lbls)
        print "done"
    print "ok, all updated."
    # update the database
    update = encoded_status.update(encoded_status.c.project_id==review_id)
    update.execute(labels_last_updated = datetime.datetime.now())

def check_encoded_status_table():
    '''
    look at all the entries in the EncodedStatus table; for any reviews
    that are present but not yet encoded, encode them.
    '''

    unencoded_review_ids = list(select([encoded_status.c.project_id], \
                                    encoded_status.c.is_encoded == False).execute())
    for unencoded_id in unencoded_review_ids:
        unencoded_id = unencoded_id.project_id
        print "encoding review %s.." % unencoded_id
        base_dir = encode_review(unencoded_id)#, base_dir="C:/dev/abstrackr_web/encode_test")
        print "done!"
        # update the record
        update = encoded_status.update(encoded_status.c.project_id == unencoded_id)
        update.execute(is_encoded = True)
        update.execute(labels_last_updated = datetime.datetime.now())
        update.execute(base_path = base_dir)


def _get_citations_for_review(review_id):
    citation_ids = list(select([citations.c.id], \
                                    citations.c.project_id == review_id).execute())
    return citation_ids

def _do_predictions_exist_for_review(review_id):
    pred_status = \
            select([prediction_status.c.project_id, prediction_status.c.predictions_exist],\
                     prediction_status.c.project_id == review_id).execute().fetchone()
              
    if pred_status is None or not pred_status.predictions_exist:
        return False
    return True

def _get_predictions_for_review(review_id):
    '''
    map citation ids to predictions (num_yes_votes)
    '''
    #preds = list(select([predictions.c.study_id, predictions.c.num_yes_votes],\
    #                predictions.c.project_id == review_id).execute())
    preds = list(select([predictions.c.study_id, predictions.c.predicted_probability],\
                    predictions.c.project_id == review_id).execute())

    preds_d = {}              
    for study_id, prob in preds:
        preds_d[study_id] = prob
    return preds_d


###
# @TODO this is redundant with the corresponding method in review.py, but
#       we need to be able to invoke this statically, hence its re-implementation
###
def _re_prioritize(review_id, sort_by_str):
    citation_ids = [cit.id for cit in _get_citations_for_review(review_id)]
    predictions_for_review = None
    if _do_predictions_exist_for_review(review_id):
        # this will be a dictionary mapping citation ids to
        # the number of yes votes for them
        predictions_for_review = _get_predictions_for_review(review_id)
    else:
        # then we have to sort randomly, since we haven't any predictions
        sort_by_str = "Random"

    # we'll map citation ids to the newly decided priorities;
    # these will depend on the sort_by_str
    cit_id_to_new_priority = {}
    if sort_by_str == "Random":
        ordering = range(len(citation_ids))
        # shuffle
        random.shuffle(ordering)
        cit_id_to_new_priority = dict(zip(citation_ids, ordering))
    elif sort_by_str == "Most likely to be relevant":
        # sort the citations by ascending likelihood of relevance
        cits_to_preds = {}
        # first insert entries for *all* citations, set this to 11
        # to prioritize newly added citations (don't want to accidently
        # de-prioritize highly relevant citations). this will take care 
        # of any citations without predictions -- 
        # e.g., a review may have been merged (?) -- citations for which
        # we have predictions will simply be overwritten, below
        for citation_id in citation_ids:
            #cits_to_preds[citation_id] = 11
            cits_to_preds[citation_id] = 1.0

        for study_id, prob in predictions_for_review.items():
            cits_to_preds[study_id] = prob
        
        # now we will sort by *descending* order; those with the most yes-votes first
        sorted_cit_ids = sorted(cits_to_preds.iteritems(), key=itemgetter(1), reverse=True)

        # now just assign priorities that reflect the ordering w.r.t. the predictions
        for i, cit in enumerate(sorted_cit_ids):
            cit_id_to_new_priority[cit[0]] = i

    #####
    #   TODO -- ambiguous case (i.e., uncertainty sampling)
    ###

    ####
    # now update the priority table for the entries
    # corresponding to this review to reflect
    # the new priorities (above)
    priority_ids_for_review = list(select([priorities.c.id, priorities.c.citation_id], \
                                    priorities.c.project_id == review_id).execute())
    for priority_id, citation_id in priority_ids_for_review:
        if citation_id in cit_id_to_new_priority:
            priority_update = \
                priorities.update(priorities.c.id == priority_id)

            priority_update.execute(priority = cit_id_to_new_priority[citation_id])
        

if __name__ == "__main__":
    print "checking the EncodedStatus table for new reviews..."
    check_encoded_status_table()
    print "done."

    print "now checking to see if I should update labels..."
    # for each review, get newest label; check this against
    # the last_updated_field

    encoded_reviews = list(select([encoded_status.c.project_id, encoded_status.c.labels_last_updated],\
                                 encoded_status.c.is_encoded==True).execute())
   
    for encoded_review in encoded_reviews:
        review_id, labels_last_updated = encoded_review

        sort_by_str = \
            select([reviews.c.sort_by], reviews.c.id == review_id).execute()
        # uh-oh
        if sort_by_str.rowcount == 0:
            print "I can't do anything for review %s -- it doesn't appear to have an entry" % review_id
        else:
            sort_by_str = sort_by_str.fetchone().sort_by
            labels_for_review = select([labels.c.label_last_updated], \
			labels.c.project_id==review_id).order_by(labels.c.label_last_updated.desc()).execute()
            if labels_for_review.rowcount > 0:
                print "checking review %s..." % review_id
                most_recent_label = labels_for_review.fetchone().label_last_updated
                print "the most recent label for review %s is dated: %s" % (review_id, most_recent_label)
                if most_recent_label > labels_last_updated:
                    # then there's been at least one new label, update encoded files!
                    print "updating labels for review %s..." % review_id
                    update_labels(review_id)
            
                    # now make predictions for updated reviews.
                    print "making predictions"
                    make_predictions(review_id)

                    # now re-prioritize
                    print "re-prioritizing..."
                    _re_prioritize(review_id, sort_by_str)
                else:
                    # initial set of predictions
                    if not _do_predictions_exist_for_review(review_id):
                        make_predictions(review_id)
                        print "ok, now re-prioritizing..."
                        _re_prioritize(review_id, sort_by_str)

    print "done."



'''
Byron C. Wallace
4/21/2014
'''

import os, pdb, pickle
import datetime
import sys

####
# external dependencies
####

import sklearn_predictor
from sklearn_predictor import BaggedUSLearner

import sqlalchemy
from sqlalchemy import *
from sqlalchemy.sql import select
from sqlalchemy.sql import and_, or_

# home-grown
import abstrackr_dataset

engine = create_engine("mysql://root@127.0.0.1:3306/abstrackr?unix_socket=/var/mysql/mysql.sock")
#engine = create_engine("mysql://abstrackr-user:5xl.z=Uy6d@127.0.0.1:3306/abstrackrDBP01?unix_socket=/var/mysql/mysql.sock")
conn = engine.connect()
metadata = MetaData(bind=engine)

### 
# table binding.
#encoded_status = Table("EncodedStatuses", metadata, autoload=True) # we no longer need this.
prediction_status = Table("PredictionStatuses", metadata, autoload=True)
predictions_table = Table("Predictions", metadata, autoload=True)
citations = Table("Citations", metadata, autoload=True)
priorities = Table("priorities", metadata, autoload=True)
labels = Table("Labels", metadata, autoload=True)
reviews = Table("Projects", metadata, autoload=True)
users = Table("user", metadata, autoload=True)
labeled_features = Table("labeledfeatures", metadata, autoload=True)
#encoded_status = Table("encodedstatuses", metadata, autoload=True)


def make_predictions(review_id):
    #predictions, train_size, num_pos = pred_results 
    print "making predictions for review: %s" % review_id
    ids, titles, abstracts, mesh, lbls_dict = get_data_for_review(review_id)
    review_dataset = abstrackr_dataset.Dataset(ids, titles, abstracts, mesh, 
                                                lbls_dict, name=str(review_id))
    learner = BaggedUSLearner(review_dataset)
    print "training..."
    learner.train()
    print "ok! making predictions..."
    pred_d, train_size, num_pos = learner.predict_remaining()
    print "-"*20 + " summary for review %s " % review_id + "-"*20
    print "training set size: %s\ntest set size: %s\n# predicted to be positive: %s" % (
            train_size, len(pred_d), num_pos)
    print "-"*62
    print "done. updating tables..."
   
    _update_predictions(review_id, pred_d, train_size, num_pos)
    print "okey dokey.\n"

def _update_predictions(review_id, predictions, train_size, num_pos):
    ####
    # update the database
    ####

    # first, delete all prediction entries associated with this
    # review (these are presumably 'stale' now)
    conn.execute(
        predictions_table.delete().where(predictions_table.c.project_id == review_id))

    # map -1's to 0's; this is because the ORM (sql-alchemy)
    # expects boolean fields to be either 0 or 1, which is apparently
    # a new thing in later releases (oh, well).
    neg_one_to_0 = lambda x : 0 if x < 0 else x

    # now re-insert them, reflecting the new prediction
    for study_id, pred_d in predictions.items():
        conn.execute(predictions_table.insert().values(study_id=study_id, project_id=review_id, \
                    prediction=neg_one_to_0(pred_d["prediction"]), num_yes_votes=pred_d["num_yes_votes"]),
                    predicted_probability=pred_d["pred_prob"])
    
    # delete any existing prediction status entries, should they exist
    conn.execute(prediction_status.delete().where(prediction_status.c.project_id == review_id))

    # finally, update the prediction status
    conn.execute(prediction_status.insert().values(project_id=review_id, predictions_exist=True,\
         predictions_last_made=datetime.datetime.now(), train_set_size=train_size,\
         num_pos_train=num_pos))


def get_data_for_review(review_id):
    '''
    ids, titles, abstracts, mesh, lbl_dict
    '''
    lbls_dict = _get_lbl_d_for_review(review_id)
    review_citation_dict = _get_ti_ab_mh(review_id)
    # note the ordering
    ids = sorted(review_citation_dict.keys())
    titles, abstracts, mesh = [], [], []
    for id_ in ids:
        citation_d = review_citation_dict[id_]
        titles.append(citation_d["title"])
        abstracts.append(citation_d["abstract"])
        mesh.append(citation_d["keywords"])
    return ids, titles, abstracts, mesh, lbls_dict


def _get_ti_ab_mh(review_id):
    fields = ["title", "abstract", "keywords"]
    none_to_text= lambda x: "none" if x is None else x

    s = citations.select(citations.c.project_id==review_id)
    citations_for_review = list(s.execute())
    cit_d = {}
    for citation in citations_for_review:
        citation_id = citation['id']
        cit_d[citation_id] = {}
        for field in fields:
            cit_d[citation_id][field] = citation[field]
    return cit_d

def _get_lbl_d_for_review(review_id):
    lbl_d = {}
    s = labels.select(labels.c.project_id==review_id)
    for lbl in s.execute():
        lbl_d[lbl["study_id"]]=lbl["label"]  
    return lbl_d




try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

setup(
    name='abstrackr',
    version='0.3.0',
    description='Web-based Citation Screening Tool',
    author='Byron Wallace',
    author_email='byron_wallace@brown.edu',
    url='https://research.brown.edu/myresearch/Byron_Wallace',
    install_requires=[
        "Pylons==1.0",
        "SQLAlchemy>=0.5",
        "TurboMail>=3.0.3,<=3.0.99",
        "repoze.what-quickstart>=1.0.8,<=1.0.99",
        "repoze.what_pylons>=1.0,<=1.0.99",
        "MySQL-python>=1.2.3,<=1.2.99",
        "elementtree>=1.2.7,<=1.2.99",
        "WebTest>=1.2.3,<=1.2.99",
        "WebOb>=1.0.7,<=1.0.99",
    ],
    setup_requires=["PasteScript>=1.6.3"],
    packages=find_packages(exclude=['ez_setup']),
    include_package_data=True,
    test_suite='nose.collector',
    package_data={'abstrackr': ['i18n/*/LC_MESSAGES/*.mo']},
    # message_extractors={'abstrackr': [
    #        ('**.py', 'python', None),
    #        ('templates/**.mako', 'mako', {'input_encoding': 'utf-8'}),
    #        ('public/**', 'ignore', None)]},
    zip_safe=False,
    paster_plugins=['PasteScript', 'Pylons'],
    entry_points="""
    [paste.app_factory]
    main = abstrackr.config.middleware:make_app

    [paste.app_install]
    main = pylons.util:PylonsInstaller
    """,
)



import pdb
from collections import defaultdict

import sklearn
from sklearn import svm
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.grid_search import GridSearchCV
# sklearn.svm.LinearSVC
import numpy as np

class MetaClf:

    def __init__(self, clf_ti, clf_ab, clf_mh=None):
        self.clf_ti = clf_ti
        self.clf_ab = clf_ab
        self.clf_mh = clf_mh

    def predict(self, X_ti, X_ab, X_mh=None):
        preds_ti = self.clf_ti.predict(X_ti)
        preds_ab = self.clf_ab.predict(X_ab)
        if self.clf_mh is None or X_mh is None:
            preds_mh = [1]*X_ti.shape[0]
        else:
            preds_mh = self.clf_mh.predict(X_mh)
        # at least one is -1 here; not sure if you want that 
        # or 0
        return preds_ti + preds_ab + preds_mh >= -1

    def predict_probabilities(self, X_ti, X_ab, X_mh):
        probs_ti = self.clf_ti.predict_proba(X_ti)[:,1]
        probs_ab = self.clf_ab.predict_proba(X_ab)[:,1]
        if self.clf_mh is None or X_mh is None:
            probs_mh = np.asarray([.5]*X_ti.shape[0])
        else:
            probs_mh = self.clf_mh.predict_proba(X_mh)[:,1]
        # @TODO should probably do something clever here
        # (e.g., calibrate the weightings apportioned to 
        #  each view)
        return (probs_ti + probs_ab + probs_mh) / 3.0

class BaggedUSLearner:
    """ bagged & undersampled learner. """
    def __init__(self, dataset, ensemble_size=11):
        self.dataset = dataset
        self.ensemble_size = ensemble_size

    def train(self):
        X_titles, X_abstracts, X_mesh, y = self.dataset.get_train_X_y()
        if max(y) < 0:
            raise Exception, "no positive examples!"

        print "training on %s labeled instances" % len(y)

        self.ensemble = []

        param_grid = [{'C': [.001, .01, 1, 10, 100]}]
        
        ###
        # undersampling
        pos_indices = np.where(y==1)[0]
        n_pos = len(pos_indices)
        neg_indices = np.where(y==-1)[0]
        for l_i in xrange(self.ensemble_size):
            neg_indices_i = np.random.choice(neg_indices, n_pos)
            indices_i = np.hstack((pos_indices, neg_indices_i))
        
            X_i_ti, X_i_ab = X_titles[indices_i], X_abstracts[indices_i]
            if X_mesh is not None:
                X_i_mh = X_mesh[indices_i]

            y_i = y[indices_i]

            ###
            # might want to swap in a different scoring function here,
            # but on the other hand, this is balanced accuracy, so probably
            # what we want (i guess?). could also try recall..
            self.clf_ti = GridSearchCV(svm.SVC(probability=True, kernel="linear"), param_grid)
            self.clf_ti.fit(X_i_ti, y_i)

            self.clf_ab = GridSearchCV(svm.SVC(probability=True, kernel="linear"), param_grid)
            self.clf_ab.fit(X_i_ab, y_i)

            self.clf_mh = None
            if X_mesh is not None:
                self.clf_mh = GridSearchCV(svm.SVC(probability=True, kernel="linear"), param_grid)
                self.clf_mh.fit(X_i_mh, y_i)

            self.ensemble.append(MetaClf(self.clf_ti, self.clf_ab, self.clf_mh))

        print "ok -- all trained (ensemble built)"
        
    #### TODO this needs to be a dictionary with study id's
    # as keys and values should themselves be dictionaries
    # with 
    def predict_remaining(self, return_true_labels=False):
        #test_X, test_y = self.dataset.get_test_X_y()
        X_titles, X_abstracts, X_mesh, y = self.dataset.get_test_X_y()
        ids = self.dataset.get_test_ids()

        predictions = []
        predicted_probabilities = []
        for ensemble in self.ensemble:
            predictions.append(ensemble.predict(X_titles, X_abstracts, X_mesh).tolist())
            predicted_probabilities.append(ensemble.predict_probabilities(X_titles, X_abstracts, X_mesh).tolist())

        final_preds_d = defaultdict(dict)
        k = float(self.ensemble_size)/2
        num_pos = 0
        for i in xrange(X_titles.shape[0]):
            id_ = ids[i]
            ensemble_preds_i = [preds[i] for preds in predictions]
            pos_count = ensemble_preds_i.count(True)
            final_prediction_i = pos_count >= k
            if final_prediction_i:
                num_pos += 1
            #final_predictions.append(final_prediction_i)
            final_preds_d[id_]["prediction"] = final_prediction_i
            final_preds_d[id_]["num_yes_votes"] = pos_count
            ### 
            # @TODO reimplement our ICDM stuff!
            # these are really awful.
            predicted_probs_i = [pred_probs[i] for pred_probs in predicted_probabilities]
            final_predicted_prob_i = sum(predicted_probs_i) / float(self.ensemble_size)

            final_preds_d[id_]["pred_prob"] = final_predicted_prob_i

        train_size = self.dataset.N - len(ids)
        return final_preds_d, train_size, num_pos


"""Setup the abstrackr application"""
import logging
import os.path

import pylons.test

from abstrackr.config.environment import load_environment
from abstrackr.model import meta
from abstrackr.model.meta import Session, Base

# for authentication
from abstrackr.model import User, Group, Permission
log = logging.getLogger(__name__)

def setup_app(command, conf, vars):
    """Place any commands to setup abstrackr here"""
    load_environment(conf.global_conf, conf.local_conf)

    filename = os.path.split(conf.filename)[-1]
    if filename == 'test.ini':
        # Permanently drop any existing tables
        log.info("Dropping existing tables...")
        Base.metadata.drop_all(checkfirst=True, bind=Session.bind)
    # Create the tables if they don't already exist
    log.info("Creating database tables")
    Base.metadata.drop_all(checkfirst=True, bind=Session.bind)
    Base.metadata.create_all(bind=Session.bind)
    log.info("Finished setting up")

    g = Group()
    g.name = u'admin'
    meta.Session.add(g)
    p = Permission()
    p.name = u'admin'
    meta.Session.add(p)

    # give myslf a login. 
    u = User()
    u.username = u'byron'
    u.fullname = u'byron wallace'
    u.experience = 2
    u._set_password('pignic')
    u.email = u'byron.wallace@gmail.com'
    meta.Session.add(u)
    meta.Session.commit()



'''
a slightly misnamed module, as it can actually handle more than XML.
basically, this contains code to map data from a given format
(pmid list, refman xml, etc.) to abstrackr's SQL representation.
'''

# std libs
import pdb
import string 
import csv
import random
import os

# third party
import sqlite3
import elementtree
from elementtree.ElementTree import ElementTree
import Bio
from Bio import Entrez

# homegrown 
import pubmedpy
import abstrackr.model as model # for access to sqlalchemy

pubmedpy.set_email("byron.wallace@gmail.com")

# for tsv importing 
OBLIGATORY_FIELDS = ["id", "title", "abstract"] # these have to be there
OPTIONAL_FIELDS = ["authors", "keywords", "journal", "pmid"]

MAX_TITLE_LENGTH = 480
START_FILE_MARKER = "\xef\xbb\xbf" # AKA BOM

def looks_like_tsv(file_path):
    header_line = open(file_path, 'rU').readline()
    headers = [x.lower().strip().replace(START_FILE_MARKER, "") for x in header_line.split("\t")]
    if len(headers) == 0:
        return False
    
    # title, etc.?
    if all([_field_in(field, headers) for field in OBLIGATORY_FIELDS]):
        return True
    
    return False

def _field_in(field, headers):
    # we'll only enforce that a string *start* with
    # the field; this avoids issues with, e.g.,
    # pluralization and new lines.
    return any([x.startswith(field) for x in headers])

def _parse_pmids(pmids_path):
    parsing_errors = []
    pmids = []
    for pmid in [x.strip() for x in open(pmids_path, 'rU').readlines()]:
        if pmid != "":
            try:
                pmids.append(int(pmid))
            except:
                error = "Unable to parse '%s' in '%s'. This does not appear to be a valid Pubmed ID" % (pmid, os.path.basename(pmids_path))
                parsing_errors.append(error)
    return pmids, {"import-errors": parsing_errors}


def pubmed_ids_to_d(pmids):
    print "fetching articles..."
    articles = pubmedpy.batch_fetch(pmids)
    print "ok."

    articles = [article for article in articles if len(article)>=3]
    
    pmids_d = {}
    none_to_str = lambda x: x if x is not None else ""
    
    for article in articles:
        title_text = article.get("TI")
        ab_text = article.get("AB")    
        authors = none_to_str(article.get("AU"))
        journal = none_to_str(article.get("JT"))
        keywords = none_to_str(article.get("MH"))
        pmid = int(article["PMID"]) 
        pmids_d[pmid] = {"title":title_text, "abstract":ab_text, "journal":journal,\
                                             "keywords":keywords, "pmid":pmid, "authors":authors}
    return pmids_d
                                
def pmid_list_to_sql(pmids_path, review):
    pmids, dict_misc = _parse_pmids(pmids_path)
    d = pubmed_ids_to_d(pmids)
    print "ok. now inserting into sql..."
    dict_to_sql(d, review)
    print "ok."
    return len(d), dict_misc

def ris_to_sql(ris_path, review):
    print "building a dictionary from %s..." % ris_path
    d = ris_to_d(open(ris_path).readlines())
    print "ok. now inserting into sql..."
    dict_to_sql(d, review)
    print "ok."
    return len(d)
    
def ris_to_d(ris_data):
    ris_d = {}
    cur_id = 1
    cur_authors, cur_keywords = [], []
    current_citation = {"title":"", "abstract":"", "journal":"",\
                                "keywords":"", "pmid":"", "authors":""}

    # drop garbage/blank lines
    ris_data = [line for line in ris_data if "-" in line]

    # we skip the first line which just starts the
    # first citation (citation 1)
    for line in ris_data[1:]:
        field, value = line.split("-")[0], "-".join(line.split("-")[1:])
        field, value = field.strip(), value.strip()

        if field == "TY":
            # new citation
            current_citation["authors"] = list(set(cur_authors))
            current_citation["keywords"] = list(cur_keywords)
            ris_d[cur_id] = current_citation

            ###
            # now create a new (empty) citation
            # to be overwritten
            current_citation = {"title":"", "abstract":"", "journal":"",\
                                "keywords":"", "pmid":"", "authors":""}
            cur_authors, cur_keywords = [], []
            cur_id += 1
        elif field in ("AU", "A1"):
            # author
            #pdb.set_trace()
            cur_authors.append(value)
        elif field in ("T1", "TI"):
            current_citation["title"] = value[:MAX_TITLE_LENGTH]
        elif field.startswith("J"):
            current_citation["journal"] = value
        elif field == "KW":
            cur_keywords.append(value)
        elif field in ("N2", "AB"):
            current_citation["abstract"] = value
    # add the last citation
    ris_d[cur_id] = current_citation
    
    return ris_d


def tsv_to_d(citations, field_index_d):
    tsv_d = {}
    parsing_errors = []

    for citation in citations:
        ## Make sure that the fields we are interested in are within range first
        ## If the line is misformed then we don't want to even consider when committing
        ## any data to the database and so we skip to the next citation without adding
        ## anything to tsv_d dictionary but we record the error for reporting
        try:
            for field in OPTIONAL_FIELDS:
                if field in field_index_d:
                    citation[field_index_d[field]].decode('utf8', 'replace')
            for field in OBLIGATORY_FIELDS:
                citation[field_index_d[field]].decode('utf8', 'replace')
        except (IndexError, KeyError) as e:
            error = "Problem with %s. This line is misformed and missing critical fields" % (citation)
            parsing_errors.append(error)
            continue
        ## Now build the tsv_d dictionary. If we did this earlier then we
        ## will get into trouble for having an entry with missing fields and it
        ## will throw errors later in the program
        cur_id = citation[field_index_d["id"]]
        tsv_d[cur_id] = {}
        for field in OPTIONAL_FIELDS:
            if field in field_index_d:
                tsv_d[cur_id][field] = \
                    citation[field_index_d[field]].decode('utf8', 'replace')
    
                # issue 2 -- if this is the authors field, we expect author names
                # to be separated by commas. later in the pipeline, we'll expect
                # a *list* here, so we create that now.
                if field == "authors" or field=="keywords":
                    tsv_d[cur_id][field] = tsv_d[cur_id][field].split(",")
            else:
                # just insert a blank string
                tsv_d[cur_id][field] = ""
    
        # now add the obligatory fields
        for field in OBLIGATORY_FIELDS:
            tsv_d[cur_id][field] = citation[field_index_d[field]].decode('utf8', 'replace')
    return tsv_d, {"import-errors": parsing_errors}


def tsv_to_sql(tsv_path, review):
    # figure out the indices
    #citations = open(tsv_path).readlines()
    open_f = open(tsv_path, 'rU')
    citations = csv.reader(open_f, delimiter="\t")
    # map field names to the corresponding indices
    # in the tsv, as indicated by the header
    headers = [header.lower().strip().replace(START_FILE_MARKER, "") for header in citations.next()]
    field_index_d = _field_index_d(headers)
    d, dict_misc = tsv_to_d(citations, field_index_d)
    dict_to_sql(d, review)
    open_f.close()
    return len(d), dict_misc

#def ris_to_d(ris_data):
#    ris_d = {}
#    cur_id = 1
#    cur_authors, cur_keywords = [], []
#    current_citation = {"title":"", "abstract":"", "journal":"",\
#                                "keywords":"", "pmid":"", "authors":""}
#
#    # drop garbage/blank lines
#    ris_data = [line for line in ris_data if "-" in line]
#
#    # we skip the first line which just starts the
#    # first citation (citation 1)
#    for line in ris_data[1:]:
#        field, value = line.split("-")[0], "-".join(line.split("-")[1:])
#        field, value = field.strip(), value.strip()
#
#        if field == "TY":
#            # new citation
#            current_citation["authors"] = list(set(cur_authors))
#            current_citation["keywords"] = list(cur_keywords)
#            ris_d[cur_id] = current_citation
#
#            ###
#            # now create a new (empty) citation
#            # to be overwritten
#            current_citation = {"title":"", "abstract":"", "journal":"",\
#                                "keywords":"", "pmid":"", "authors":""}
#            cur_authors, cur_keywords = [], []
#            cur_id += 1
#        elif field in ("AU", "A1"):
#            # author
#            #pdb.set_trace()
#            cur_authors.append(value)
#        elif field in ("T1", "TI"):
#            current_citation["title"] = value[:MAX_TITLE_LENGTH]
#        elif field.startswith("J"):
#            current_citation["journal"] = value
#        elif field == "KW":
#            cur_keywords.append(value)
#        elif field in ("N2", "AB"):
#            current_citation["abstract"] = value
#    # add the last citation
#    ris_d[cur_id] = current_citation
#    
#    return ris_d
    
def _field_index_d(headers):
    field_index_d = {}
    for field in OBLIGATORY_FIELDS:
        # we know these exist
        field_index_d[field] = headers.index(field)
    
    # now let's see if they've optional headers
    for field in OPTIONAL_FIELDS:
        if field in headers:
            field_index_d[field] = headers.index(field)
    
    return field_index_d

def xml_to_sql(xml_path, review):
    print "building a dictionary from %s..." % xml_path
    d, dict_misc = xml_to_dict(xml_path)
    print "ok. now inserting into sql..."
    dict_to_sql(d, review)
    print "ok."
    return len(d), dict_misc


    
def dict_to_sql(xml_d, review): 
    cit_num = 0
    # issue #31: explicitly randomize ordering
    xml_d_items = xml_d.items()
    random.shuffle(xml_d_items)
    for ref_id, citation_d in xml_d_items:
        cit_id = insert_citation(review.id, ref_id, citation_d)
        insert_priority_entry(review.id, cit_id, cit_num)
        cit_num += 1
    model.Session.commit()

def insert_citation(review_id, ref_id, citation_d):
    citation = model.Citation()
    citation.project_id = review_id
    try:
        ref_id = int(ref_id)
    except:
        ref_id = None

    citation.refman = ref_id
    pmid = citation_d['pmid']
    citation.pmid =  pmid if (pmid is not None and pmid != '') else 0
    # we truncate the citation if it's too long!
    citation.title = citation_d['title'][:MAX_TITLE_LENGTH] if \
                                citation_d['title'] is not None else "(no title found)"
    citation.abstract = citation_d['abstract'][:9980] if \
                                citation_d['abstract'] is not None else None
    citation.authors = " and ".join(citation_d['authors'])
    citation.keywords = ','.join(citation_d['keywords'])
    citation.journal = citation_d['journal']
    
    model.Session.add(citation)
    model.Session.commit()

    return citation.id
 
def insert_priority_entry(review_id, citation_id, \
                            init_priority_num, num_times_labeled=0):
    priority = model.Priority()
    priority.project_id = review_id
    priority.citation_id = citation_id
    priority.priority = init_priority_num
    priority.num_times_labeled = num_times_labeled
    priority.is_out = False
    model.Session.add(priority)
    

def xml_to_dict(fpath):
    '''
    Converts study data from (ref man generated) XML to a dictionary matching study IDs (keys) to
    title/abstract tuples (values). For example: dict[n] might map to a tuple [t_n, a_n] where t_n is the
    title of the nth paper and a_n is the abstract
    '''
    ref_ids_to_abs = {}
    parsing_errors = []
    num_no_abs = 0
    tree = ElementTree(file=fpath)
    
    num_failed = 0
    
    for record in tree.findall('.//record'):
        pubmed_id, refmanid = None, None

        refman_version = record.findtext('.//source-app')
        path_str = None
        ### here we check the RefMan version, and change
        # the xml path accordingly. this fixes issue #7
        if refman_version == 'Reference Manager 12.0':
            path_str = './/rec-number/style'
            journal_path_str = './/periodical/full-title/style'
        elif refman_version == 'Reference Manager 11.0':
            path_str = './/rec-number'
            journal_path_str = './/periodical/abbr-1/style'

        try:
            refmanid = int(record.findtext(path_str))
        except:
            error = "Unable to parse record '%s' in '%s'" % (record, os.path.basename(fpath))
            #print "failed to parse refman document"
            parsing_errors.append(error)

        if refmanid is not None:
            # attempt to grab the pubmed id
            pubmed_id = ""
            try:
                pubmed = record.findtext('.//notes/style')
                pubmed = pubmed.split("-")
                for i in range(len(pubmed)):
                    if "UI" in pubmed[i]:
                        pubmed_str = pubmed[i+1].strip()
                        pubmed_id = int("".join([x for x in pubmed_str if x in string.digits]))
            except Exception, ex:
                error = "Problem getting pmid from '%s' in '%s'" % (record, os.path.basename(fpath))
                parsing_errors.append(error)
                #print "problem getting pmid ..."
                #print ex
                #print("\n")
    
            ab_text = record.findtext('.//abstract/style')
            if ab_text is None:
                num_no_abs += 1
    
            title_text = record.findtext('.//titles/title/style')
    
            # Also grab keywords
            keywords = [keyword.text.strip().lower() for keyword in record.findall(".//keywords/keyword/style")]

            # and authors
            authors = [author.text for author in record.findall(".//contributors/authors/author/style")]

            # journal
            journal = record.findtext(journal_path_str)

            ref_ids_to_abs[refmanid] = {"title":title_text, "abstract":ab_text, "journal":journal,\
                        "keywords":keywords, "pmid":pubmed_id, "authors":authors}

    
    print "\nFinished. Returning %s title/abstract/keyword sets, %s of which have no abstracts.\n" \
                    % (len(ref_ids_to_abs.keys()), num_no_abs)
    return ref_ids_to_abs, {"import-errors": parsing_errors}




